//
// Title:   AffineRegistration.cpp
// Created: March 1, 2003
// Authors: Daniel Cotting, Oliver Knoll, Mark Pauly, Matthias Zwicker
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include <gsl/gsl_multifit.h>
#include <qstringlist.h>
#include <qmessagebox.h>

#include "../../../../Core/Interfaces/src/FrameBufferInterface.h"
#include "../../../../Core/Interfaces/src/SurfelInterface.h"
#include "../../../../Core/Configuration/src/Configuration.h"
#include "../../../../Core/RendererManager/src/RendererManager.h"
#include "../../../../Core/CoreTools/src/CoreTools.h"
#include "../../../../Core/DataStructures/src/SurfelCollection.h"
#include "../../../../Core/CoreResourceTools/src/CoreResourceTools.h"
#include "../../../../Core/Brush/src/Brush.h"
#include "../../../../Core/Scene/src/Scene.h"
#include "../../../../Core/Painter/src/Painter.h"
#include "../../../../Core/MarkerManager/src/MarkerManager.h"
#include "../../../../Core/MarkerManager/src/MarkerInterface.h"
#include "../../../../Core/MarkerManager/src/Marker3D.h"
#include "../../../../Utilities/src/ImageOps.h"
#include "../../../../Utilities/src/Common.h"
#include "../../../../Utilities/UserMessages/src/StatusBar.h"
#include "AffineRegistration.h"

// **************
// public methods
// **************

AffineRegistration::AffineRegistration() {

	// get informed whenever the selection has changed
	this->connect (CoreTools::getInstance()->getSelectionTool(), SIGNAL (selectionChanged (SurfelInterface::Flag)),
		           this, SLOT (handleSelectionChanged (SurfelInterface::Flag)));
	
}

AffineRegistration::~AffineRegistration() {
}

void AffineRegistration::registerObjects(int firstObjectID, int currentObjectID, QList<MarkerInterface>* markers1, QList<MarkerInterface>* markers2) {

	if (markers1->count()<4) {
		QMessageBox::information( NULL, "Registration Error", "At least 4 markers per object have to be defined." );
		return;
	}
	if (markers1->count()!=markers2->count()) {
		QMessageBox::information( NULL, "Registration Error", "Marker counts of both object do not match." );
		return;
	}
	switch( QMessageBox::information( NULL, "Question",	"Are you sure that the verification object is activated?",
								"Yes", "No", 0, 0, 1 ))	{
	case 0:
		break;
	case 1: 
		return;
	}

	Object* verificationObject = Scene::getInstance()->getActiveObject();
	QList<MarkerInterface>* verificationObjectMarkers;
	QList<MarkerInterface>* originalObjectMarkers;
	if (firstObjectID == currentObjectID) {
		verificationObjectMarkers = markers1;
		originalObjectMarkers = markers2;
	} else {
		verificationObjectMarkers = markers2;
		originalObjectMarkers = markers1;
	}

	StatusBar *statusBar = StatusBar::getInstance();
	statusBar->showMessage ("Allocating data structures...");
	
	gsl_matrix *X, *cov;
	gsl_vector *y1, *y2, *y3, *y4, *w, *c;

	int n = verificationObjectMarkers->count();

	X  = gsl_matrix_alloc (n, 4);
	y1 = gsl_vector_alloc (n);
	y2 = gsl_vector_alloc (n);
	y3 = gsl_vector_alloc (n);
	y4 = gsl_vector_alloc (n);
	w  = gsl_vector_alloc (n);
	c = gsl_vector_alloc (4);

	cov = gsl_matrix_alloc (4, 4);
	double chisq;

	statusBar->showMessage ("Filling data structures...");
	Vector3D verificationPosition = ((Marker3D*)verificationObjectMarkers->first())->getSurfel()->getPosition();
	Vector3D originalPosition = ((Marker3D*)originalObjectMarkers->first())->getSurfel()->getPosition();
	for (int i = 0; i < n; i++) {
		gsl_matrix_set (X, i, 0, verificationPosition[0]);
		gsl_matrix_set (X, i, 1, verificationPosition[1]);
		gsl_matrix_set (X, i, 2, verificationPosition[2]);
		gsl_matrix_set (X, i, 3, 1);
		gsl_vector_set (y1, i, originalPosition[0]);
		gsl_vector_set (y2, i, originalPosition[1]);
		gsl_vector_set (y3, i, originalPosition[2]);
		gsl_vector_set (y4, i, 1);
		Marker3D* verificationMarker = (Marker3D*)verificationObjectMarkers->next();
		Marker3D* originalMarker = (Marker3D*)originalObjectMarkers->next();
		if (verificationMarker!=NULL) {
			verificationPosition = verificationMarker->getSurfel()->getPosition();
			originalPosition = originalMarker->getSurfel()->getPosition();
		}
    }

	statusBar->showMessage ("Solving least squares equation...");
    gsl_multifit_linear_workspace * work = gsl_multifit_linear_alloc (n, 4);
    gsl_multifit_linear (X, y1, c, cov, &chisq, work);
	float m11 = gsl_vector_get(c,0);
	float m12 = gsl_vector_get(c,1);
	float m13 = gsl_vector_get(c,2);
	float m14 = gsl_vector_get(c,3);
    gsl_multifit_linear (X, y2, c, cov, &chisq, work);
	float m21 = gsl_vector_get(c,0);
	float m22 = gsl_vector_get(c,1);
	float m23 = gsl_vector_get(c,2);
	float m24 = gsl_vector_get(c,3);
    gsl_multifit_linear (X, y3, c, cov, &chisq, work);
	float m31 = gsl_vector_get(c,0);
	float m32 = gsl_vector_get(c,1);
	float m33 = gsl_vector_get(c,2);
	float m34 = gsl_vector_get(c,3);
    gsl_multifit_linear (X, y4, c, cov, &chisq, work);
	float m41 = gsl_vector_get(c,0);
	float m42 = gsl_vector_get(c,1);
	float m43 = gsl_vector_get(c,2);
	float m44 = gsl_vector_get(c,3);
    gsl_multifit_linear_free (work);

	char message[200];
	sprintf(message,	"Registration matrix:\n"
						" %03.4f %03.4f %03.4f %03.4f\n"
						" %03.4f %03.4f %03.4f %03.4f\n"
						" %03.4f %03.4f %03.4f %03.4f\n"
						" %03.4f %03.4f %03.4f %03.4f\n", 
						m11, m12, m13, m14,
						m21, m22, m23, m24,
						m31, m32, m33, m34,
						m41, m42, m43, m44);

	QMessageBox::information( NULL, "Information", message);

	statusBar->showMessage ("Transforming points...");
	SurfelCollection* activeSurfelCollection = verificationObject->getSurfelCollection();
	int nofSurfels = activeSurfelCollection->getNofSurfels();
	SurfelInterface* surfel;
	Scene* scene = Scene::getInstance();

	for (i = 0, surfel = activeSurfelCollection->getFirstSurfel(); i < nofSurfels; i++, surfel = activeSurfelCollection->getNextSurfel()) {
		Vector3D position = surfel->getPosition();
		float w = m41*position[0]+m42*position[1]+m43*position[2]+m44;
		float x = (m11*position[0]+m12*position[1]+m13*position[2]+m14)/w;
		float y = (m21*position[0]+m22*position[1]+m23*position[2]+m24)/w;
		float z = (m31*position[0]+m32*position[1]+m33*position[2]+m34)/w;
		position[0] = x;
		position[1] = y;
		position[2] = z;
		scene->setSurfelPosition(surfel, position, false);
	}

	// emit update
	scene->setSurfelPosition(activeSurfelCollection->getFirstSurfel(), activeSurfelCollection->getFirstSurfel()->getPosition(), true);

	scene->emitObjectModifiedSignal (true);
	scene->emitRenderFrameSignal ();
	statusBar->showMessage ("Registration done.");
}


// *************
// private slots
// *************

void AffineRegistration::handleSelectionChanged (SurfelInterface::Flag selectionFlag) {
	
	if (selectionFlag == SurfelInterface::SELECTED1) {
	}
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
