//
// Title:   AffineRegistrationTool.cpp
// Created: March 1, 2003
// Authors: Daniel Cotting, Oliver Knoll, Mark Pauly, Matthias Zwicker
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include <limits.h>

#include <qtoolbar.h>
#include <qtoolbutton.h>
#include <qevent.h>
#include <qpixmap.h>
#include <qwhatsthis.h>
#include <qwidget.h>
#include <qstring.h>
#include <qpainter.h>
#include <qlist.h>
#include <qfont.h>

#include "../../../../Core/Interfaces/src/RendererInterface.h"
#include "../../../../Core/RendererManager/src/RendererManager.h"
#include "../../../../Core/Scene/src/Scene.h"
#include "../../../../Core/MarkerManager/src/Marker2D.h"
#include "../../../../Core/MarkerManager/src/Marker3D.h" 
#include "../../../../Core/MarkerManager/src/MarkerUtils.h"
#include "../../../../Core/MarkerManager/src/MarkerManager.h"
#include "AffineRegistrationDialogImpl.h"
#include "AffineRegistrationTool.h"

// ******************
// 'What's this' text
// ******************

static const char *whatsThisText = "<img source=\"affineregistration\"> "
"Registers two objects by undoing an affine transformation.";

// ************** 
// public methods
// **************

AffineRegistrationTool::AffineRegistrationTool() {

	AffineRegistrationDialogImpl *affineRegistrationDialog;

	toolCursor          = new QCursor (CrossCursor);
	configurationWidget = new AffineRegistrationConfigurationWidget();
	affineRegistration    = new AffineRegistration();

	markers1 = new QList<MarkerInterface>;
	markers2 = new QList<MarkerInterface>;

	affineRegistrationDialog = AffineRegistrationDialogImpl::getInstance (this);

	// store initial position and size of the dialog, as to restore it for the next time the dialog is opened
	affineRegistrationDialogPosition = affineRegistrationDialog->pos();
	affineRegistrationDialogSize     = affineRegistrationDialog->size();

	isDragging = false;
	firstObjectID = -1;
}

AffineRegistrationTool::~AffineRegistrationTool() {
	delete toolCursor;
	delete configurationWidget;
	delete affineRegistration;

	AffineRegistrationDialogImpl::destroyInstance();
}

QToolButton *AffineRegistrationTool::addButtonToToolBar (QToolBar *toolBar) {
	
	QIconSet iconSet;
	QPixmap  normalIcon ("Resources/AffineRegistrationToolIcon.png"),
		     disabledIcon ("Resources/AffineRegistrationToolIcon_disabled.png");

	iconSet.setPixmap (normalIcon, QIconSet::Automatic, QIconSet::Normal);
	iconSet.setPixmap (disabledIcon, QIconSet::Automatic, QIconSet::Disabled);
	affineRegistrationButton = new QToolButton (iconSet, "Affine Registration Tool", "Affine Registration",
	                                          0, 0, toolBar, "Affine Registration Tool");
	affineRegistrationButton->setToggleButton (true);
	affineRegistrationButton->setEnabled (this->checkToolAvailability());
	QWhatsThis::add (affineRegistrationButton, whatsThisText);
	QMimeSourceFactory::defaultFactory()->setPixmap ("affineregistration", normalIcon);

	return affineRegistrationButton;

}

QToolButton *AffineRegistrationTool::getToolButton() {
	return affineRegistrationButton;
}

QCursor *AffineRegistrationTool::getToolCursor() const {
	return toolCursor;
}

QWidget *AffineRegistrationTool::getToolConfigurationWidget() const {
	return configurationWidget;
}

bool AffineRegistrationTool::isAvailable() const {
	return this->checkToolAvailability();
}

void AffineRegistrationTool::handleMousePressEvent (QMouseEvent *me) {

	QWidget       *rendererWidget;

	// NOTE: no need to call renderFrame, since the markers call this method themselves to guarantee
	// that the frame buffer contains the necessary information (i.e., per pixel surfel lists)

	if (me->button() == QMouseEvent::LeftButton) {

		// **********************************
		// modify the fitting constraint list
		// **********************************
		int objectID = Scene::getInstance()->getActiveObjectID();
		if (firstObjectID < 0) {
			firstObjectID = objectID;
		}
		bool isFirstObject = (firstObjectID==objectID);
		QList<MarkerInterface>* markers;
		if (isFirstObject) {
			markers = markers1;
		} else {
			markers = markers2;
		}

		rendererWidget = RendererManager::getInstance()->getMainRenderer()->getRendererWidget();
		selectedMarker = (Marker3D *)MarkerUtils::pickMarker (me->pos(), markers, rendererWidget);
		if (selectedMarker == 0) {
			MarkerUtils::deselectAll (markers, rendererWidget);
		}

	}

}

void AffineRegistrationTool::handleMouseMoveEvent (QMouseEvent *me) {

	QWidget *rendererWidget;

	if (me->state() == QMouseEvent::LeftButton) {

		if (selectedMarker != 0) {
			isDragging = true;
			// from now on this marker ignores external calls to 'togglePainting', e.g. when
			// the render widget updates its contents
			selectedMarker->setDragged (true);
			rendererWidget = RendererManager::getInstance()->getMainRenderer()->getRendererWidget();
			keepMarker = selectedMarker->moveTo (me->pos(), rendererWidget);
		}

	}

}

void AffineRegistrationTool::handleMouseReleaseEvent (QMouseEvent *me) {

	Marker3D        *marker;
	SurfelInterface *surfel;
	QWidget         *rendererWidget;
	
	rendererWidget = RendererManager::getInstance()->getMainRenderer()->getRendererWidget();

	int objectID = Scene::getInstance()->getActiveObjectID();
	if (firstObjectID < 0) {
		firstObjectID = objectID;
	}
	bool isFirstObject = (firstObjectID==objectID);
	QList<MarkerInterface>* markers;
	if (isFirstObject) {
		markers = markers1;
	} else {
		markers = markers2;
	}

	if (me->button() == QMouseEvent::LeftButton) {

		// **********************************
		// modify the fitting constraint list
		// **********************************

		if (selectedMarker == 0 && isDragging == false) {

			// *****************
			// create new marker
			// *****************

			surfel = MarkerUtils::getClosestSurfel (me->pos());
			if (surfel != 0) {

				marker = new Marker3D (surfel, true, markers->count()+1, MarkerInterface::FITTING_CONSTRAINT_STYLE);
				marker->togglePainting (rendererWidget);
				markers->append(marker);
			}

		} else if (selectedMarker != 0 && isDragging == false) {

			// ****************
			// toggle selection
			// ****************
			MarkerUtils::toggleSelection (selectedMarker, rendererWidget);

		} else {

			// ********************
			// check dragged marker
			// ********************
			if (keepMarker == false) {
				markers->remove(selectedMarker);
			} else {
				selectedMarker->setDragged (false);
			}
			isDragging = false;

		}

	} else if (me->button() == QMouseEvent::RightButton) {
		MarkerUtils::deselectAll (markers, rendererWidget);
	}

}

void AffineRegistrationTool::handleKeyPressEvent (QKeyEvent *ke) {
	ke->ignore();		
}

void AffineRegistrationTool::handleKeyReleaseEvent (QKeyEvent *ke) {
	ke->ignore();		
}

void AffineRegistrationTool::removeSelectedMarkers() {

	MarkerInterface *marker3D;
	bool            hasListChanged;

	QWidget         *rendererWidget;
	rendererWidget = RendererManager::getInstance()->getMainRenderer()->getRendererWidget();

	QList<MarkerInterface>* markers;
	markers = markers1;
	marker3D = markers->first();
	hasListChanged = false;
	while (marker3D != 0) {

		if (marker3D->isSelected() == true) {

			marker3D->togglePainting (rendererWidget);
			// remove the current item, the current item is automatically the next item
			// note: if the last item is removed
			//       - in Qt 2.x: the new last item (the one just before the deleted item) becomes the 
			//                    new current item, if there is one
			//       - in Qt 2.x: the new current item is set to 0
			//       -> in Qt 2.x the second last item might be checked twice
			markers->remove();
			marker3D = markers->current();
			hasListChanged = true;

		}
		else {
			// check the next item
			marker3D = markers->next();
		}

	}

	// check if list has changed, re-enumerate and notify event receivers
	if (hasListChanged == true) {	
		MarkerUtils::reallocateIDs (markers1, rendererWidget);
	}

	markers = markers2;
	marker3D = markers->first();
	hasListChanged = false;
	while (marker3D != 0) {

		if (marker3D->isSelected() == true) {

			marker3D->togglePainting (rendererWidget);
			// remove the current item, the current item is automatically the next item
			// note: if the last item is removed
			//       - in Qt 2.x: the new last item (the one just before the deleted item) becomes the 
			//                    new current item, if there is one
			//       - in Qt 2.x: the new current item is set to 0
			//       -> in Qt 2.x the second last item might be checked twice
			markers->remove();
			marker3D = markers->current();
			hasListChanged = true;

		}
		else {
			// check the next item
			marker3D = markers->next();
		}

	}

	// check if list has changed, re-enumerate and notify event receivers
	if (hasListChanged == true) {	
		MarkerUtils::reallocateIDs (markers2, rendererWidget);
	}
}

void AffineRegistrationTool::toggleMarkers() {
	QWidget         *rendererWidget;
	rendererWidget = RendererManager::getInstance()->getMainRenderer()->getRendererWidget();
	MarkerInterface *marker3D;

	for (marker3D = markers1->first(); marker3D != 0; marker3D = markers1->next()) {
		marker3D->togglePainting (rendererWidget);
	}
	for (marker3D = markers2->first(); marker3D != 0; marker3D = markers2->next()) {
		marker3D->togglePainting (rendererWidget);
	}
}

void AffineRegistrationTool::removeAllMarkers() {
	toggleMarkers();
	markers1->clear();
	markers2->clear();
	firstObjectID = -1;
}

void AffineRegistrationTool::registerObjects() {
	int currentObjectID = Scene::getInstance()->getActiveObjectID();
	affineRegistration->registerObjects(firstObjectID, currentObjectID, markers1, markers2);
}


void AffineRegistrationTool::selectTool() {

	AffineRegistrationDialogImpl *affineRegistrationDialog;
	
	affineRegistrationDialog = AffineRegistrationDialogImpl::getInstance (this);

	// this method of restoring a QWidget's position is according to the Qt documentation:
	// "This method works on both MS-Windows and most existing X11 window managers"
	affineRegistrationDialog->resize (affineRegistrationDialogSize);
	affineRegistrationDialog->move (affineRegistrationDialogPosition);
	affineRegistrationDialog->show();

	toggleMarkers();
}

void AffineRegistrationTool::deselectTool() {

	AffineRegistrationDialogImpl	*affineRegistrationDialog;

	affineRegistrationDialog = AffineRegistrationDialogImpl::getInstance (this);

	// store current position and size of the dialog, as to restore it for the next time the dialog is opened
	affineRegistrationDialogPosition = affineRegistrationDialog->pos();
	affineRegistrationDialogSize     = affineRegistrationDialog->size();

	AffineRegistrationDialogImpl::destroyInstance();

	toggleMarkers();
}

// ************
// public slots
// ************

void AffineRegistrationTool::handleRendererWidgetRepainted() {
	toggleMarkers();
}

// ***************
// private methods
// ***************

bool AffineRegistrationTool::checkToolAvailability() const {
	return RendererManager::getInstance()->getMainRenderer()->isInteractive() &&
		   (Scene::getInstance()->getNofObjects() > 0);
}

// ****************
// access functions
// ****************

ToolInterface *createTool() {
	return new AffineRegistrationTool();
}

const char *getToolName() {
	return "Affine Registration Tool";
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
