//
// Title:   LaplaceFrequencyTransformation.h 
// Created: November 28, 2002
// Authors: Daniel Cotting (for adaptation to PointShop3D, main authors see below)
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.

/**
 * @file LaplaceFrequencyTransformation.h
 *
 * Contains the laplace frequency transformation for surfels.
 */

#ifndef _LAPLACEFREQUENCYTRANSFORMATION_H_EA6C0EE0_BC30_11d5_B625_A58C4DF45B22_INCLUDED
#define _LAPLACEFREQUENCYTRANSFORMATION_H_EA6C0EE0_BC30_11d5_B625_A58C4DF45B22_INCLUDED

///////////////////////////////////////////////////////////////////////////////////////
// INCLUDES
///////////////////////////////////////////////////////////////////////////////////////

#include <gsl/gsl_matrix_float.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_eigen.h>

#include "MySurfelList.h"
#include "FrequencyTransformationInterface.h"
#include "WatermarkInterface.h"
#include "GenericVector.h"
#include "PCA.h"

#include "../../../Core/Scene/src/Scene.h"
#include "../../../Core/DataStructures/src/NeighbourHood.h"
#include "../../../Utilities/UserMessages/src/StatusBar.h"

// number of entries that are non-zero in the laplace-matrix
static uint nofLaplaceMatrixNonZeroEntries;
static float maxValence;

// BEGIN: the following is used for the LAPACK version of eigenproblem
	//extern "C"
	//{
	//#include "blaswrap.h"
	//#include "f2c.h"
	//#include "clapack.h"
	//}
// END: the following is used for the LAPACK version of eigenproblem

// BEGIN: the following is used for the ARPACK version of eigenproblem
#include "dsaupd.h"
static double **laplaceMatrixArpack = NULL;
static void av(uint n, double *in, double *out)
{
	uint i;
	for (i=0; i<n; i++) out[i] = 0;
	for (i=0; i<nofLaplaceMatrixNonZeroEntries; i++) out[(int)laplaceMatrixArpack[i][0]] += in[(int)laplaceMatrixArpack[i][1]] * laplaceMatrixArpack[i][2];
}
// END: the following is used for the ARPACK version of eigenproblem


// BEGIN: the following is used for the MESCHACH version of eigenproblem
extern "C" {
 #include	"matrix2.h"
 #include	"sparse2.h"
 #include	"iter.h"
}
static SPMAT* laplaceMatrixMeschach;
// END: the following is used for the MESCHACH version of eigenproblem

///////////////////////////////////////////////////////////////////////////////////////
// CLASS DECLARATION
///////////////////////////////////////////////////////////////////////////////////////

/**
 * This template class computes a laplace frequency transformation on a surfel list.
 *
 * Template parameters: <tt>nofAttributes</tt>: The number of attributes to embed watermark into.
 *
 * Usually the number of attributes is set to 3. If it is set to a value larger than 5
 * both positions and colors are considered. To embed information only into the color 
 * attributes set the number to 3 and pass a color watermark to the constructor.
 */
template <unsigned nofAttributes>
class LaplaceFrequencyTransformation: public FrequencyTransformationInterface<nofAttributes>
{

private:

	/**
     * Eigensystem of the matrix will be stored in a linked list of elements of structure eigenRec.
	 */
	struct eigenRec {
		char *annotation;				///< annotation of the entry in the eigensystem list
		int cols;						///<
		double eigenvalue;				///< eigenvalue 
		gsl_vector_float *eigenvector;	///< eigenvector
		struct eigenRec *next;			///< pointer to the next entry ein the list
	};

	/** 
	 * These typedefs will allow us to easily use the linked list representing
	 * the datalist and eigensystem throughout our code.
	 */	
	typedef struct eigenRec EigenRec;
	typedef EigenRec* EigenList;

public:

	/**
	 * Contains the data structure necessary for the backward transformation.
	 */
	class LaplaceDataPack: public FrequencyTransformationInterface<nofAttributes>::DataPack {
	public:
		/**
		 * Constructor for the class LaplaceDataPack (empty).
		 */
		LaplaceDataPack() {};

		/**
		 * Destructor for the class LaplaceDataPack.
		 */
		~LaplaceDataPack() {
			EigenList eigenHead = eigenList;
			eigenList=eigenList->next;
			while (eigenList!=NULL) {
				EigenList nextElement = eigenList->next;
				free(eigenList->annotation);
				gsl_vector_float_free(eigenList->eigenvector);
				free(eigenList);
				eigenList = nextElement;
			}
			free(eigenHead);
			assert(remainingOffset!=NULL);
			delete[] remainingOffset;
			remainingOffset=NULL;
		}

		/**
		 * Contains the list of eigenvectors. Required as basis for backward transformation.
		 */
		EigenList eigenList;

		/**
		 * Offset required to reconstruct initial data during backward transformation.
		 * We need this because we are not calculating all the eigenvalues and eigenvectors.
		 * The offset can be seen as a combination of all the remaining eigenvectors.
		 */
		GenericVector<nofAttributes>* remainingOffset;
	};

	/**
	 * Constructor of the class <tt>LaplaceFrequencyTransformation</tt>.
	 *
	 * @param watermark The watermark which will be embedded later on. Required to get the size of watermark.
	 */
	LaplaceFrequencyTransformation(WatermarkInterface* watermark);

	/**
	 * Performs the forward transformation using the laplace matrix.
	 *
	 * @param surfelList The surfels to transform into frequency domain.
	 * @param spectralListOut The <tt>SpectralList</tt> object which will store the result.
	 * @param dataPackOut The <tt>DataPack</tt> object which will store temporary data required for backward transformation.
	 */
	void forwardTransformation( 
		MySurfelList* surfelList, 
		SpectralList** spectralListOut,
		DataPack** dataPackOut);

	/**
	 * Performs a fast version of the forward transformation using the laplace matrix (does skip some
	 * steps not necessary if the topology of the points is the same as in a previous calculation).
	 *
	 * @param surfelList		The surfels to transform into frequency domain.
	 * @param dataPackIn		The <tt>DataPack</tt> object which already stores information 
	 *							about a calculation performed on a point set with the same topology.
	 * @param spectralListOut	The <tt>SpectralList</tt> object which will store the result.
	 */
	void forwardTransformation(
			MySurfelList* surfelList, 
			DataPack* dataPackIn,
			SpectralList** spectralListOut);

	/**
	 * Performs the backward transformation using the laplace matrix.
	 *
	 * @param surfelList The list of the surfels whose positions will be changed to the values resulting from backward transformation.
	 * @param spectralList The <tt>SpectralList</tt> object containing the spectral respresentation.
	 * @param dataPack The <tt>DataPack</tt> object storing temporary data required for backward transformation.
	 */
	void backwardTransformation(
		MySurfelList* surfelList, 
		SpectralList* spectralList, 
		DataPack* dataPack);

	/**
	 * Returns the name of this module.
	 *
	 * @return The name of this module.
	 */
	const char* getModuleName() const;

	/**
	 * Returns the description of this module.
	 *
	 * @return The description of this module.
	 */
	const char* getModuleDescription() const;

private:

	/**
	 * Creates the laplace matrix of the surfel attributes specified as parameter.
	 *
	 * @param surfelAttributes An array of attribute vectors.
	 * @param dimension The dimension of the attribute array.
	 * @return The Laplace matrix.
	 */
	gsl_matrix_float* createLaplaceMatrix(
		GenericVector<nofAttributes>* surfelAttributes, 
		uint dimension);

	/**
	 * Returns the eigensystem of the laplace matrix.
	 *
	 * @param laplaceMatrix The matrix to calculate the eigensystem for.
	 * @return The list containing the eigenvalues and eigenvectors.
	 */
	EigenList calculateEigenSystem(
		gsl_matrix_float* laplaceMatrix);
	
	/**
	 * Normalizes the eigenvectors in the list passed to this function.
	 *
	 * @param eigenList The list containing the eigensystem whose eigenvectors have to be normalized.
	 */
	void normalizeEigenVectors(
		EigenList eigenList);
	
	/**
	 * Projects the surfel attributes onto the eigensystem.
	 *
	 * @param surfelAttributes The surfel attributes to project onto the eigensystem.
	 * @param eigenList The eigensystem to project the surfel attributes onto.
	 * @return The resulting spectral representation.
	 */
	SpectralList* project(
		GenericVector<nofAttributes>* surfelAttributes,
		EigenList eigenList);

	/**
	 * Add a list element to the specified list.
	 *
	 * @param annotation The annotation of the list element to append to the list.
	 * @param eigenList The head of the list we want to add the element to.
	 */
	void addEigenLink( 
		char* annotation, 
		EigenList eigenList );

	/**
	 * Initializes the list which will contain the eigensystem.
	 *
	 * @param cols The number of eigenvectors/eigenvalues this list will host.
	 * @return The list containing the eigensystem.
	 */
	EigenList initializeEigenList(
		int cols);

	/** 
	 * Gets the remaining offset, that is the difference between the correct result of 
	 * the backward transformation and the data we get when considering the few spectral
	 * values we have calculated. In other words the remaining offset corresponds to the
	 * part that is contributed by the spectral values we have not computed.
	 *
	 * @param dimension Dimension of the surfel attribute array.
	 * @param surfelAttributes The array containing the surfel attribute vectors.
	 * @param specralList The spectral list.
	 * @param LaplaceDataPack The data pack containing the temporary data for backward transformation.
	 * @return The remaining offset.
	 */
    GenericVector<nofAttributes>* getRemainingOffset(
		uint dimension, 
		GenericVector<nofAttributes>* surfelAttributes, 
		SpectralList* spectralList,
		LaplaceDataPack* laplaceDataPack);

	/**
	 * Number of eigenvalues that have to be calculated.
	 */
	uint nofEigenvalues;

	/**
	 * Defines whether the watermark has to be applied to the color or the position attributes.
	 */
	bool isColorWatermark;

	/**
	 * Defines whether the watermark has to be applied by using the fast embedding algorithm.
	 */
	bool useFastVersion;

};

///////////////////////////////////////////////////////////////////////////////////////
// CLASS DEFINITION
///////////////////////////////////////////////////////////////////////////////////////

/**
 * Constructor of the class <tt>LaplaceFrequencyTransformation</tt>.
 *
 * @param watermark The watermark which will be embedded later on. Required to get the size of watermark.
 */
template <unsigned nofAttributes>
LaplaceFrequencyTransformation<nofAttributes>::LaplaceFrequencyTransformation(
		WatermarkInterface* watermark) 
{
	nofEigenvalues   = watermark->getUsedSizeInBits();
	isColorWatermark = watermark->isColorWatermark();
	useFastVersion   = watermark->getUseFastVersion();
}
 
/**
 * Performs the forward transformation using the laplace matrix.
 *
 * @param surfelList The surfels to transform into frequency domain.
 * @param spectralListOut The <tt>SpectralList</tt> object which will store the result.
 * @param dataPackOut The <tt>DataPack</tt> object which will store temporary data required for backward transformation.
 */
template <unsigned nofAttributes>
void LaplaceFrequencyTransformation<nofAttributes>::forwardTransformation(
		MySurfelList* surfelList, 
		SpectralList** spectralListOut,
		DataPack** dataPackOut)
{
	
	uint i;
	StatusBar* statusBar = StatusBar::getInstance();
	 
	// create surfel attributes data structure
	statusBar->showMessage ("FT: Surfel Attributes...", 0);
	uint dimension = surfelList->count();
	GenericVector<nofAttributes>* surfelAttributes = new GenericVector<nofAttributes>[dimension];

	SurfelInterface* surfel;
	for (i = 0, surfel = surfelList->first(); i < dimension; i++, surfel = surfelList->next()) {
		// generic vector can store any attributes
		// this allows a watermark to be embedded not only in the positions of the surfels
		// but also in other attributes like color and texture properties
		// note that the neighborhood relationship will only be computed using the positions
		Vector3D position = surfel->getPosition();
		// set the position attributes
		surfelAttributes[i][0] = position[0];
		surfelAttributes[i][1] = position[1];
		surfelAttributes[i][2] = position[2];
		// here we could add some additional attributes like color...
		// ....
		// ....
		// ....
		if (nofAttributes>5) {
			// add three color components
			QColor color = surfel->getDiffuseColor();
			surfelAttributes[i][3] = color.red();
			surfelAttributes[i][4] = color.green();
			surfelAttributes[i][5] = color.blue();
		}
	}
	
	// create the laplacian matrix based on neighbourhood
	statusBar->showMessage ("FT: Laplace Matrix...", 0);
	gsl_matrix_float* laplaceMatrix = createLaplaceMatrix(surfelAttributes, dimension);
	
	// calculate the normalized eigenvectors of the system
	statusBar->showMessage ("FT: Eigensystem...", 0);
	LaplaceDataPack* laplaceDataPack = new LaplaceDataPack();
	laplaceDataPack->eigenList = calculateEigenSystem(laplaceMatrix);

	statusBar->showMessage ("FT: Normalize...", 0);
	normalizeEigenVectors(laplaceDataPack->eigenList);

	// the user wants to substitute the color for the position
	// so let's overwrite the previously written information
	if (isColorWatermark) {
		for (i = 0, surfel = surfelList->first(); i < dimension; i++, surfel = surfelList->next()) {
			QColor color = surfel->getDiffuseColor();
			surfelAttributes[i][0] = color.red();
			surfelAttributes[i][1] = color.green();
			surfelAttributes[i][2] = color.blue();
		}
	}

	// project each component of the vertex coordinates onto the eigenvectors
	statusBar->showMessage ("FT: Project...", 0);
	SpectralList* spectralList = project(surfelAttributes, laplaceDataPack->eigenList);

	// get remaining offset
	statusBar->showMessage ("FT: Rem. Offset...", 0);
	laplaceDataPack->remainingOffset = getRemainingOffset(dimension, surfelAttributes, spectralList, laplaceDataPack);

	// return the spectral representation
	*spectralListOut = spectralList;
	
	// return the temporary data that is required for the inverse transformation
	*dataPackOut = (DataPack*)laplaceDataPack;
	
	// free unused memory
	delete[] surfelAttributes;
	gsl_matrix_float_free(laplaceMatrix);
}

/**
 * Performs a fast version of the forward transformation using the laplace matrix (does skip some
 * steps not necessary if the topology of the points is the same as in a previous calculation).
 *
 * @param surfelList		The surfels to transform into frequency domain.
 * @param dataPackIn		The <tt>DataPack</tt> object which already stores some information 
 *							about a calculation performed on a point set with the same topology.
 * @param spectralListOut	The <tt>SpectralList</tt> object which will store the result.
 */
template <unsigned nofAttributes>
void LaplaceFrequencyTransformation<nofAttributes>::forwardTransformation(
		MySurfelList* surfelList, 
		DataPack* dataPackIn,
		SpectralList** spectralListOut)
{
	
	uint i;
	StatusBar* statusBar = StatusBar::getInstance();
	 
	// create surfel attributes data structure
	statusBar->showMessage ("FT: Surfel Attributes...", 0);
	uint dimension = surfelList->count();
	GenericVector<nofAttributes>* surfelAttributes = new GenericVector<nofAttributes>[dimension];

	SurfelInterface* surfel;
	for (i = 0, surfel = surfelList->first(); i < dimension; i++, surfel = surfelList->next()) {
		// generic vector can store any attributes
		// this allows a watermark to be embedded not only in the positions of the surfels
		// but also in other attributes like color and texture properties
		// note that the neighborhood relationship will only be computed using the positions
		Vector3D position = surfel->getPosition();
		// set the position attributes
		surfelAttributes[i][0] = position[0];
		surfelAttributes[i][1] = position[1];
		surfelAttributes[i][2] = position[2];
		// here we could add some additional attributes like color...
		// ....
		// ....
		// ....
		if (nofAttributes>5) {
			// add three color components
			QColor color = surfel->getDiffuseColor();
			surfelAttributes[i][3] = color.red();
			surfelAttributes[i][4] = color.green();
			surfelAttributes[i][5] = color.blue();
		}
		// the user wants to substitute the color for the position
		// so let's overwrite the previously written information
		if (isColorWatermark) {
			QColor color = surfel->getDiffuseColor();
			surfelAttributes[i][0] = color.red();
			surfelAttributes[i][1] = color.green();
			surfelAttributes[i][2] = color.blue();
		}
	}
	
	// project each component of the vertex coordinates onto the eigenvectors
	statusBar->showMessage ("FT: Project...", 0);
	SpectralList* spectralList = project(surfelAttributes, ((LaplaceDataPack*)dataPackIn)->eigenList);

	// return the spectral representation
	*spectralListOut = spectralList;
	
	// free unused memory
	delete[] surfelAttributes;
}	



/** 
 * Gets the remaining offset, that is the difference between the correct result of 
 * the backward transformation and the data we get when considering the few spectral
 * values we have calculated. In other words the remaining offset corresponds to the
 * part that is contributed by the spectral values we have not computed.
 *
 * @param dimension Dimension of the surfel attribute array.
 * @param surfelAttributes The array containing the surfel attribute vectors.
 * @param specralList The spectral list.
 * @param LaplaceDataPack The data pack containing the temporary data for backward transformation.
 * @return The remaining offset.
 */
template <unsigned nofAttributes>
GenericVector<nofAttributes>* LaplaceFrequencyTransformation<nofAttributes>::getRemainingOffset(
		uint dimension, 
		GenericVector<nofAttributes>* surfelAttributes,
		SpectralList* spectralList,
		LaplaceDataPack* laplaceDataPack) 
{
	uint i;
	uint j;

	// get the eigenlist
	EigenList eList = laplaceDataPack->eigenList;
	// skip the head of the list
	eList = eList->next; 
	
	// initialize the components
	gsl_vector_float* components[nofAttributes];
	for (i=0; i<nofAttributes; i++) {
			components[i] = gsl_vector_float_calloc(dimension); // calloc sets to zero
	}

	// calculate part of the backward transformation
	gsl_vector_float* temp = gsl_vector_float_alloc(dimension);
	GenericVector<nofAttributes>* spectralCoefficient;
	for (spectralCoefficient = spectralList->first(); spectralCoefficient!=0; spectralCoefficient = spectralList->next()) {
		for (i=0; i<nofAttributes; i++) {
			gsl_vector_float_memcpy(temp, eList->eigenvector);
			gsl_vector_float_scale(temp, (*spectralCoefficient)[i]);
			gsl_vector_float_add(components[i], temp);
		}
		eList = eList->next;
	}
	
	// initialize remaining offset
	GenericVector<nofAttributes>* remainingOffset = new GenericVector<nofAttributes>[dimension];

	// calculate remaining offset
	for (i=0; i<dimension; i++) {
		for (j=0; j<nofAttributes; j++) {
			remainingOffset[i][j] = surfelAttributes[i][j] - gsl_vector_float_get(components[j], i); 
		}
	}

	// free memory
	gsl_vector_float_free(temp);
	for (i=0; i<nofAttributes; i++) {
		gsl_vector_float_free(components[i]);
	}

	// return value
	return remainingOffset;
}

/**
 * Performs the backward transformation using the laplace matrix.
 *
 * @param surfelList The list of the surfels whose positions will be changed to the values resulting from backward transformation.
 * @param spectralList The <tt>SpectralList</tt> object containing the spectral respresentation.
 * @param dataPack The <tt>DataPack</tt> object storing temporary data required for backward transformation.
 */
template <unsigned nofAttributes>
void LaplaceFrequencyTransformation<nofAttributes>::backwardTransformation(
		MySurfelList* surfelList, 
		SpectralList* spectralList, 
		DataPack* dataPack) 
{

	uint i;
	uint j;

	// get the eigenlist
	EigenList eList = ((LaplaceDataPack*)dataPack)->eigenList;
	// skip the head of the list
	eList = eList->next; /* Skip the head of the list */

	// initialize the components
	unsigned int dimension = surfelList->count();
	unsigned int nofSurfelAttributes = nofAttributes;
	gsl_vector_float* components[nofAttributes];
	for (i=0; i<nofSurfelAttributes; i++) {
			components[i] = gsl_vector_float_calloc(dimension); // calloc sets to zero
	}

	// calculate part of the backward transformation
	GenericVector<nofAttributes>* spectralCoefficient;
	gsl_vector_float* temp = gsl_vector_float_alloc(dimension);
	for (spectralCoefficient = spectralList->first(); spectralCoefficient!=0; spectralCoefficient = spectralList->next()) {
		for (i=0; i<nofSurfelAttributes; i++) {
			gsl_vector_float_memcpy(temp, eList->eigenvector);
			gsl_vector_float_scale(temp, (*spectralCoefficient)[i]);
			gsl_vector_float_add(components[i], temp);
		}
		eList = eList->next;
	}
	
	// modify position data in surfel list
	GenericVector<nofAttributes>* remainingOffset = ((LaplaceDataPack*)dataPack)->remainingOffset;
	Scene* scene = Scene::getInstance();
	SurfelInterface* surfel;
	for (i=0, surfel = surfelList->first(); surfel!=0; i++, surfel = surfelList->next()) {

		// get the position
		Vector3D position = surfel->getPosition();
		
		// we always have at least three attributes which correspond to the location of the surfels.
		// this is only true if the boolean value isColorWatermark is not set to true.
		if (!isColorWatermark) {

			// user does not want to substitute the position with color, so let's move the positions
			for (j=0; j<3; j++) {
				position[j] = gsl_vector_float_get(components[j], i) + remainingOffset[i][j]; 
			}
			scene->setSurfelPosition(surfel, position, false);
		} else {
			
			// user does not want to change the position
			// so change color
			QColor color = surfel->getDiffuseColor();
			int red   = (int)(gsl_vector_float_get(components[0], i) + remainingOffset[i][0]);
			int green = (int)(gsl_vector_float_get(components[1], i) + remainingOffset[i][1]);
			int blue  = (int)(gsl_vector_float_get(components[2], i) + remainingOffset[i][2]);
			// clamp values
			red   = red>255?255:red;
			green = green>255?255:green;
			blue  = blue>255?255:blue;
			red   = red<0?0:red;
			green = green<0?0:green;
			blue  = blue<0?0:blue;
			color.setRgb(red, green, blue);
			scene->setSurfelDiffuseColor(surfel, color.rgb(), false);
		}

		// check if we have more than five attributes -> color is also contained (RGB)
		if (nofSurfelAttributes>5) {

			// for the time being only color can be used as an additional attribute
			QColor color = surfel->getDiffuseColor();
			int red   = (int)(gsl_vector_float_get(components[3], i) + remainingOffset[i][3]);
			int green = (int)(gsl_vector_float_get(components[4], i) + remainingOffset[i][4]);
			int blue  = (int)(gsl_vector_float_get(components[5], i) + remainingOffset[i][5]);
			// clamp values
			red   = red>255?255:red;
			green = green>255?255:green;
			blue  = blue>255?255:blue;
			red   = red<0?0:red;
			green = green<0?0:green;
			blue  = blue<0?0:blue;
			color.setRgb(red, green, blue);
			scene->setSurfelDiffuseColor(surfel, color.rgb(), false);
		}

	}

	// update screen by sending a signal
	scene->setSurfelPosition(surfelList->first(), surfelList->first()->getPosition(), true);
		
	// free memory
	gsl_vector_float_free(temp);
	for (i=0; i<nofSurfelAttributes; i++) {
		gsl_vector_float_free(components[i]);
	}
}

/**
 * Creates the laplace matrix of the surfel attributes specified as parameter.
 *
 * @param surfelAttributes An array of attribute vectors.
 * @param dimension The dimension of the attribute array.
 * @return The Laplace matrix.
 */
template <unsigned nofAttributes>
gsl_matrix_float* LaplaceFrequencyTransformation<nofAttributes>::createLaplaceMatrix(
		GenericVector<nofAttributes>* surfelAttributes, 
		uint dimension) 
{

	// allocate the memory for the laplace matrix
	// make sure the matrix is initialized to zero
	StatusBar::getInstance()->showMessage ("FT: Laplace Matrix...", 0);
	gsl_matrix_float *laplaceMatrix = gsl_matrix_float_calloc(dimension, dimension);
	
	// create data structures for the nearest neighbourhood search
	// note that generic vector inherits from class Vector3D which is used to extract the positions
	// while calculating the neighbourhood relationship
	NeighbourHood neighbourHood(surfelAttributes, dimension);
	
	// search for the 6 nearest neighbours (note that due to the fact that we have to restore
	// symmetry properties more than 6 final neighbours can result)
	uint nofNeighbours = 6;
	neighbourHood.setNofNeighbours(nofNeighbours);

	// count the number of entries in the matrix which are non-zero
	// update the global variables
	nofLaplaceMatrixNonZeroEntries = 0;
	maxValence = 0;
	// iterate over the surfels and create laplace matrix based on neighbourhood 
	for (uint surfelIndex = 0; surfelIndex < dimension; surfelIndex++) {
	
		// process the neighbours
		neighbourHood.setSourcePoint(surfelAttributes[surfelIndex]);
		for (uint j=0; j<nofNeighbours; j++) {
			uint neighbourSurfelIndex = neighbourHood.getNeighbourPositionIndex(j);
			if (neighbourSurfelIndex != surfelIndex) {
				if (gsl_matrix_float_get(laplaceMatrix, neighbourSurfelIndex, surfelIndex)==0) {
					gsl_matrix_float_set(laplaceMatrix, neighbourSurfelIndex, surfelIndex, -1);
					nofLaplaceMatrixNonZeroEntries++;
					// increment degree/valence of point
					float newValence = gsl_matrix_float_get(laplaceMatrix, neighbourSurfelIndex, neighbourSurfelIndex)+1;
					gsl_matrix_float_set(laplaceMatrix, neighbourSurfelIndex, neighbourSurfelIndex, newValence);
					if (newValence>maxValence) {
						maxValence = newValence;
					}
				}
				// also set the following position to minus one because of symmetry
				if (gsl_matrix_float_get(laplaceMatrix, surfelIndex, neighbourSurfelIndex)==0) {
					gsl_matrix_float_set(laplaceMatrix, surfelIndex, neighbourSurfelIndex, -1);
					nofLaplaceMatrixNonZeroEntries++;
					// increment degree/valence of point
					float newValence = gsl_matrix_float_get(laplaceMatrix, surfelIndex, surfelIndex)+1;
					gsl_matrix_float_set(laplaceMatrix, surfelIndex, surfelIndex, newValence);
					if (newValence>maxValence) {
						maxValence = newValence;
					}
				}
			}
		}
	} 

	// the total number of entries in the matrix is the sum of the counted items and the number of diagonal elements.
	nofLaplaceMatrixNonZeroEntries += dimension;

	// output some information about the matrix structure
	// qDebug("Sparsity: %d / %d; Dimension: %d\r\n", nofLaplaceMatrixNonZeroEntries, dimension*dimension, dimension);

	// free mem
	return laplaceMatrix;
}

// struct used for the sorting procedure
struct eigenpack {
	Real eigenvalue;
	int eigenvectorIndex;
};

/* eigenpack_compare -- comparison function to pass to qsort() */
static	int	eigenpack_compare(const eigenpack* x, const eigenpack* y) {
	Real tmp;
	tmp = (*x).eigenvalue - (*y).eigenvalue;
	return (tmp > 0 ? 1 : tmp < 0 ? -1: 0);
}

/**
 * Returns the eigensystem of the laplace matrix. The current implementation uses the Arnoldi method.
 *
 * @param laplaceMatrix The matrix to calculate the eigensystem for.
 * @return The list containing the eigenvalues and eigenvectors.
 */
template <unsigned nofAttributes>
LaplaceFrequencyTransformation<nofAttributes>::EigenList 
LaplaceFrequencyTransformation<nofAttributes>::calculateEigenSystem(gsl_matrix_float* laplaceMatrix) 
{

	uint i, j;

	StatusBar* statusBar = StatusBar::getInstance();
	statusBar->showMessage ("FT: EV Arnoldi Calculation...", 0);

	// get dimension of the matrix
	uint n = laplaceMatrix->size1;
	float* eigenValues = new float[nofEigenvalues];

	// we provide two different versions of the algorithm: a fast one and a more accurate one
	if (!useFastVersion) {

		// allocate memory for the sparse storage of the matrix
		laplaceMatrixArpack = new double*[nofLaplaceMatrixNonZeroEntries];
		for (i=0; i<nofLaplaceMatrixNonZeroEntries; i++) laplaceMatrixArpack[i] = new double[3];

		// create sparse matrix representation for arnoldi method
		uint index = 0;
		for (i=0; i<n; i++) {
			for (j=0; j<n; j++) {
				if (gsl_matrix_float_get(laplaceMatrix, i, j)!=0) {
					laplaceMatrixArpack[index][0] = i;
					laplaceMatrixArpack[index][1] = j;
					laplaceMatrixArpack[index][2] = gsl_matrix_float_get(laplaceMatrix, i, j);
					index++;
				}
			}
		}
		// tlen now contains the number of elements
		// create the data structures which will contain the eigenvalues and teh eigenvectors
		double **eigenVectorsArnoldi, *eigenValuesArnoldi;
 		eigenValuesArnoldi = new double[nofEigenvalues];
		eigenVectorsArnoldi = new double*[n];
		for (i=0; i<n; i++) eigenVectorsArnoldi[i] = new double[nofEigenvalues];

		// make the fortran call
		dsaupd(&av, n, nofEigenvalues, eigenValuesArnoldi, eigenVectorsArnoldi);

		// Now we construct the result
		statusBar->showMessage ("FT: EV Arnoldi Result...", 0);
  		for (j=0; j<nofEigenvalues; j++) {
			for (i=0; i<n; i++) { 
				gsl_matrix_float_set(laplaceMatrix, i, j, eigenVectorsArnoldi[i][j]);
			}
			eigenValues[j] = eigenValuesArnoldi[j];
			// qDebug("eigenvalue: %f, vec[0]: %f, vec[1]: %f, vec[2]: %f", eigenValues[j], eigenVectorsArnoldi[0][j], eigenVectorsArnoldi[1][j], eigenVectorsArnoldi[2][j]);
		}

		// cleanup
		delete[] eigenValuesArnoldi;
		for (i=0; i<n; i++) delete[] eigenVectorsArnoldi[i];
		delete[] eigenVectorsArnoldi;

		// deallocate global memory
		for (i=0; i<nofLaplaceMatrixNonZeroEntries; i++) {
			delete[] laplaceMatrixArpack[i];
		}
		delete[] laplaceMatrixArpack;

	} else {

		laplaceMatrixMeschach = sp_get(n, n, maxValence+1);

		for (i=0; i<n; i++) {
			for (j=0; j<n; j++) {
				if (gsl_matrix_float_get(laplaceMatrix, i, j)!=0) {
					sp_set_val(laplaceMatrixMeschach, i, j, gsl_matrix_float_get(laplaceMatrix, i, j));
				}
			}
		}
	
		Real h_rem;
		uint m = nofEigenvalues*2;
		VEC* x0 = v_get(n);
		double nsqrt = sqrt(n);
		for (i=0; i<n; i++) {
			x0->ve[i] = 1.0/nsqrt;
		}

		MAT* Q  = m_get(m,n);
		MAT* H  = m_get(m,m);
		iter_sparnoldi(laplaceMatrixMeschach,x0,m,&h_rem,Q,H);
		VEC* a = v_get(m);
		VEC* b = v_get(m-1);
		for (i=0; i<m; i++) {
			a->ve[i] = H->me[i][i];
			if (i!=m-1) {
				b->ve[i]=H->me[i][i+1];
			}
		}

		statusBar->showMessage ("FT: EV Arnoldi Trieig...", 0);
		trieig(a,b,H);
		VEC* eigenvalues = a;
		for (j=0; j<m; j++) {
			double value = 0;
			for (i=0; i<m; i++) {
				value += H->me[i][j]*H->me[i][j];
			}
			value = sqrt(value);
			for (i=0; i<m; i++) {
				H->me[i][j]/=value;
			}
		}
 
		statusBar->showMessage ("FT: EV Arnoldi Ritz Vectors...", 0);
		MAT* QT = m_get(n,m);
		MAT* eigenvectors = m_get(n,m);
		m_transp(Q, QT);
		m_mlt(QT, H, eigenvectors);

		// create the array of eigenpacks we want to sort
		eigenpack* eigenpacks = new eigenpack[m];
		for (i=0; i<m; i++) {
			eigenpacks[i].eigenvalue       = eigenvalues->ve[i];
			eigenpacks[i].eigenvectorIndex = i;

		}

		statusBar->showMessage ("FT: EV Arnoldi Sort...", 0);
		// sort the eigenvectors and the eigenvalues in ascending order
		qsort((void *)(eigenpacks),m,sizeof(eigenpack),(int (*)(const void*, const void*))eigenpack_compare);

		// Now we construct the result
		statusBar->showMessage ("FT: EV Arnoldi Pass Result...", 0);

		// now only take the nofEigenvalues first eigenvalues and their corresponding eigenvectors
  		for (j=0; j<nofEigenvalues; j++) {
			for (i=0; i<n; i++) { 
				gsl_matrix_float_set(laplaceMatrix, i, j, eigenvectors->me[i][eigenpacks[j].eigenvectorIndex]);
			}
			eigenValues[j] = eigenpacks[j].eigenvalue;
			//qDebug("eigenvalue: %f, vec[0]: %f, vec[1]: %f, vec[2]: %f", 
			//	eigenValues[j], 
			//	eigenvectors->me[0][eigenpacks[j].eigenvectorIndex],
			//	eigenvectors->me[1][eigenpacks[j].eigenvectorIndex], 
			//	eigenvectors->me[2][eigenpacks[j].eigenvectorIndex]);
		}

		// check whether we have a degenerate patch
		// the algorithm fails on some patches, in that case all eigenvalues are 0
		if (eigenValues[nofEigenvalues-1]==0) {
			// show user warning
			QMessageBox::information( NULL, "Warning", "Found degenerate patch. Ignoring..." );
			// we simply ignore that patch for embedding by setting the eigenvectors
			// that are now stored in the laplace matrix to zero
			gsl_matrix_float_set_zero(laplaceMatrix);
		}
	    // qDebug("arnoldi done.\n");

		// clean up
	    M_FREE(QT);
	    M_FREE(H);
	    M_FREE(Q);
	    M_FREE(eigenvectors);
	    SP_FREE(laplaceMatrixMeschach);
	    delete[] eigenpacks;
	    V_FREE(a);
	    V_FREE(b);
	    V_FREE(x0);
	}

	// create the list we will return to the caller
	EigenList eList = initializeEigenList(nofEigenvalues);
	EigenList eListHeader = eList;
	
	// skip the header
	eList = eList->next;

	// assign the eigensystem list the appropriate values
	uint index = 0;
	while (eList != NULL && index < nofEigenvalues) {
		eList->eigenvalue = eigenValues[index];
		eList->cols = n;
		eList->eigenvector = gsl_vector_float_alloc(n);
		gsl_matrix_float_get_col(eList->eigenvector, laplaceMatrix, index);

		// proceed to the next element
		index = index + 1;
		eList = eList->next;
	} 

	// update status bar and clean up
	statusBar->showMessage ("FT: EV Arnoldi Result Done.", 0);
	delete[] eigenValues;
	return eListHeader;
}

///////////////////////////////////////////////////////////////////////////////
// The following code fragments contain version of the eigenvalue
// solver for other libraries. To use them you have to move them to 
// the appropriate position in the function above
// further changes might be necessary because this cod ehas not been used
// for a long time
///////////////////////////////////////////////////////////////////////////////
/*
///////////////////////////////////////////////////////////////////////////////
// LAPACK ssyevr_ version
///////////////////////////////////////////////////////////////////////////////
	integer n = laplaceMatrix->size1;
	real* matrix = laplaceMatrix->data;
	integer lda = laplaceMatrix->size1;
	real* eigenvalues = new real[n];
	real initialwork[1];
	integer lwork = -1; // first get optimal size of work space
	integer* iwork = new integer[5*n]; // first get optimal size of work space
	integer* isupz = new integer[2*n]; // first get optimal size of work space
	integer info;
	real vl = 1;
	real vu = n/2;
	integer il = vl;
	integer iu = vu;
	real abstol = 0;
	integer numFound;
	real* outputMatrix = (real*)malloc(n*n*sizeof(real));
	integer initialwork2[1];
	integer lwork2 = -1; // first get optimal size of work space

	statusBar->showMessage ("FT: EV Optimal Val...", 0);
	ssyevr_("V", "I", "U", &n, matrix, &lda, &vl, &vu, &il, &iu, &abstol, &numFound,
		eigenvalues, outputMatrix, &n, isupz, 
		initialwork, &lwork, initialwork2, &lwork2, &info);
	assert(info==0);

	lwork = initialwork[0];
	real* work = new real[lwork];

	lwork2 = initialwork2[0]; 
	integer* work2 = new integer[lwork2];

	statusBar->showMessage ("FT: EV SSYSEVR...", 0);
	ssyevr_("V", "I", "U", &n, matrix, &lda, &vl, &vu, &il, &iu, &abstol, &numFound,
		eigenvalues, outputMatrix, &n, isupz,
		work, &lwork, work2, &lwork2, &info);
	assert(info==0);
	statusBar->showMessage ("FT: EV SSYSEV DONE...", 0);
*//*
///////////////////////////////////////////////////////////////////////////////
// LAPACK ssyevr_ version 2
///////////////////////////////////////////////////////////////////////////////
	integer n = laplaceMatrix->size1;
	real* matrix = laplaceMatrix->data;
	integer lda = laplaceMatrix->size1;
	real* eigenvalues = new real[n];
	real initialwork[1];
	integer lwork = -1; // first get optimal size of work space
	integer* iwork = new integer[5*n]; // first get optimal size of work space
	integer* ifail = new integer[n]; // first get optimal size of work space
	integer info;
	real vl = 1;
	real vu = n/2;
	integer il = vl;
	integer iu = vu;
	real abstol = 0;
	integer numFound;
	real* outputMatrix = (real*)malloc(n*n*sizeof(real));

	statusBar->showMessage ("FT: EV Optimal Val...", 0);
	ssyevr_("V", "I", "U", &n, matrix, &lda, &vl, &vu, &il, &iu, &abstol, &numFound,
		eigenvalues, outputMatrix, &n, 
		initialwork, &lwork, iwork, ifail, &info);
	assert(info==0);

	lwork = initialwork[0];
	real* work = new real[lwork];


	statusBar->showMessage ("FT: EV SSYSEVD...", 0);
	ssyevr_("V", "I", "U", &n, matrix, &lda, &vl, &vu, &il, &iu, &abstol, &numFound,
		eigenvalues, outputMatrix, &n, 
		work, &lwork, iwork, ifail, &info);
	assert(info==0);
	statusBar->showMessage ("FT: EV SSYSEV DONE...", 0);
*//*
///////////////////////////////////////////////////////////////////////////////
// LAPACK ssyevd_ version
///////////////////////////////////////////////////////////////////////////////
	integer n = laplaceMatrix->size1;
	real* matrix = laplaceMatrix->data;
	integer lda = laplaceMatrix->size1;
	real* eigenvalues = new real[n];
	real initialwork[1];
	integer lwork = -1; // first get optimal size of work space
	integer initialwork2[1];
	integer lwork2 = -1; // first get optimal size of work space
	integer info;
 
	statusBar->showMessage ("FT: EV Optimal Val...", 0);
	ssyevd_("V", "U", &n, matrix, &lda, eigenvalues, initialwork, &lwork, initialwork2, &lwork2, &info);
	assert(info==0);

	lwork = initialwork[0];
	real* work = new real[lwork];

	lwork2 = initialwork2[0];
	integer* work2 = new integer[lwork2];

	statusBar->showMessage ("FT: EV SSYSEVD...", 0);
	ssyevd_("V", "U", &n, matrix, &lda, eigenvalues, work, &lwork, work2, &lwork2, &info);
	assert(info==0);
	statusBar->showMessage ("FT: EV SSYSEV DONE...", 0);
*//*
///////////////////////////////////////////////////////////////////////////////
// GSL eigen_symmv version
///////////////////////////////////////////////////////////////////////////////

	unsigned int index;
	EigenList eListHeader;
	EigenList eList;
	eListHeader = initializeEList(laplaceMatrix->size1);
	eList = eListHeader;
	
	gsl_vector *eigenValues;
	gsl_matrix *working_m, *eigenVectors;	
	gsl_eigen_symmv_workspace *w;

	eigenValues = gsl_vector_alloc( laplaceMatrix->size2 );

	working_m = gsl_matrix_alloc( laplaceMatrix->size1, laplaceMatrix->size2 );
	gsl_matrix_memcpy( working_m, laplaceMatrix );	

	eigenVectors = gsl_matrix_alloc( laplaceMatrix->size2, laplaceMatrix->size2 );
	w = gsl_eigen_symmv_alloc( laplaceMatrix->size2 );
	
	gsl_eigen_symmv( working_m, eigenValues, eigenVectors, w );
	gsl_eigen_symmv_free( w );
	
	gsl_eigen_symmv_sort( eigenValues, eigenVectors, GSL_EIGEN_SORT_VAL_DESC );

	// assign the eigensystem list to the appropriate values 

	eList = eList->next;
	index = 0;
	while ( eList != NULL && index < eigenValues->size ) {

		eList->eigenvalue = gsl_vector_get( eigenValues, index );
		eList->cols = eigenValues->size;
		eList->eigenvector = gsl_vector_alloc( eigenValues->size );
		gsl_matrix_get_col( eList->eigenvector, eigenVectors, index );

		index = index + 1;
		eList = eList->next;

	}
	return eListHeader;
}
///////////////////////////////////////////////////////////////////////////////
// End of code fragments
///////////////////////////////////////////////////////////////////////////////
*/

/**
 * Normalizes the eigenvectors in the list passed to this function.
 * Also makes sure that eigenvectors are deterministic.
 *
 * @param eigenList The list containing the eigensystem whose eigenvectors have to be normalized.
 */
template <unsigned nofAttributes>
void LaplaceFrequencyTransformation<nofAttributes>::normalizeEigenVectors(
		EigenList eList) 
{
	// skip the head of the list
	eList = eList->next; 
	
	// process all the eigenvectors
	while ( eList != NULL ) {
	
		// calculate the square of the norm of the vector
		double value = 0.0;
		for (uint i=0; i<eList->eigenvector->size; i++) {
			value += gsl_vector_float_get( eList->eigenvector, i )*gsl_vector_float_get( eList->eigenvector, i );
		}

		if (value!=0) {
			// calculate the scaling factor so that length of the vector gets normalized (length 1).
			value = 1/sqrt(value);

			// scale the eigenvector
			gsl_vector_float_scale(eList->eigenvector, value);
		}

		// search the first element of the vector that is not equal to zero
		// and make sure it is positive
		// this allows us to make sure that the eigenvectors are oriented in a 
		// certain way so that different calls to the laplace frequence transformation lead to 
		// the same results (else the spectral coefficients would sometimes have their sign flipped)
		for (i=0; i<eList->eigenvector->size; i++) {
			float element = gsl_vector_float_get(eList->eigenvector, i);
			if (element<0) {
				// the eigenvector has wrong direction, so flip it
				gsl_vector_float_scale(eList->eigenvector, -1);
				break;
			} else if (element>0) {
				// don't do anything: the eigenvector has the direction we want
				break;
			}
		}

		// go to the next eigenvector
		eList = eList->next;
	}	
}

/**
 * Projects the surfel attributes onto the eigensystem.
 *
 * @param surfelAttributes The surfel attributes to project onto the eigensystem.
 * @param eigenList The eigensystem to project the surfel attributes onto.
 * @return The resulting spectral representation.
 */
template <unsigned nofAttributes>
LaplaceFrequencyTransformation<nofAttributes>::SpectralList* 
LaplaceFrequencyTransformation<nofAttributes>::project(GenericVector<nofAttributes>* surfelAttributes, 
		EigenList eList) 
{
	uint i;
	uint j;

	// skip the head of the list
	eList = eList->next; 

	// create the spectral list which will contain the return values
	SpectralList* spectralList = new SpectralList();

	// intermediary results
	uint nofSurfelAttributes = nofAttributes;
	float values[nofAttributes];
	
	// process all eigenvectors
	while ( eList != NULL ) {

		// reset values to zero
		for (j=0; j<nofSurfelAttributes; j++) {
			values[j] = 0.0;
		}

		// multiply eigenvectors element by element with a certain attribute vector of the surfels and sum up.
		// scalar product.
		for (i=0; i<eList->eigenvector->size; i++) {
			for (j=0; j<nofSurfelAttributes; j++) {
				values[j] += surfelAttributes[i][j]*gsl_vector_float_get(eList->eigenvector, i );
			}
		}
		
		// create final spectral coefficient
		GenericVector<nofAttributes>* spectralCoefficient = new GenericVector<nofAttributes>(values);
		spectralList->append(spectralCoefficient);

		// go to the next eigenvector
		eList = eList->next;
	}	

	// return the spectral values
	return spectralList;
}

/**
 * Initializes the list which will contain the eigensystem.
 *
 * @param col The number of eigenvectors/eigenvalues this list will host.
 * @return The list containing the eigensystem.
 */
template <unsigned nofAttributes>
LaplaceFrequencyTransformation<nofAttributes>::EigenList 
LaplaceFrequencyTransformation<nofAttributes>::initializeEigenList(int col) 
{

	// allocate head element of the list
	EigenList eigenHead;
	eigenHead = (EigenList)malloc(sizeof(EigenRec));
	
	// if allocation of head element failed then exit
	if (eigenHead == NULL) {
		printf( "failure to allocate space for head of eigen system list\n" );
		exit(1);
	}
	
	eigenHead->next = NULL;

	// add the remaining elements
	for (int i = 0; i < col; i++ ) {
		addEigenLink( "unknown", eigenHead );
	}

	// return the head of the list
	return eigenHead;
}

/**
 * Add a list element to the specified list.
 *
 * @param annotation The annotation of the list element to append to the list.
 * @param eigenList The head of the list we want to add the element to.
 */
template <unsigned nofAttributes>
void LaplaceFrequencyTransformation<nofAttributes>::addEigenLink(char *annotation, EigenList eigenList) {

	// create new element
	EigenRec* newRec;
	
	// if no head element exists, then create one
	if (eigenList == NULL) {
		eigenList = (EigenList)malloc(sizeof(EigenRec));
		printf("allocating a new list\n");
	}

	// allocate memory for the new element
	newRec = (EigenList)malloc(sizeof(EigenRec));

	// output error message if this failed
	if (newRec == NULL) {
		printf("Falied to alloc space for an EIGENREC.\n");
		exit(1);
	}

	// add annotation
	newRec->next = NULL;
	newRec->annotation = (char *)malloc(strlen(annotation)+1);
	strcpy(newRec->annotation, annotation);

	// traverse list until the last element is reached
	while(eigenList->next != NULL) {
		eigenList = eigenList->next;
	}

	// attach new element to the last one
	eigenList->next = newRec;
}

/**
 * Returns the name of this module.
 *
 * @return The name of this module.
 */
template <unsigned nofAttributes, bool colorInsteadOfPosition>
const char* LaplaceFrequencyTransformation<nofAttributes>::getModuleName() const {
	return "Laplace Frequency Transformation";
}

/**
 * Returns the description of this module.
 *
 * @return The description of this module.
 */
template <unsigned nofAttributes>
const char* LaplaceFrequencyTransformation<nofAttributes>::getModuleDescription() const {
	return "Transforms attributes of surfels to a frequency domain by using the laplace transformation.";
}

#endif // _LAPLACEFREQUENCYTRANSFORMATION_H_EA6C0EE0_BC30_11d5_B625_A58C4DF45B22_INCLUDED

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
