//
// Title:   MLSProjection.cpp
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

/** 
 * @file MLSProjection.cpp
 *
 * Projects a patch of the original object onto the surface of the verification object using MLS.
 */

#include <math.h>
#include <gsl/gsl_multifit.h>

#include "../../../Core/Scene/src/Scene.h"
#include "../../../Core/DataStructures/src/NeighbourHood.h"
#include "../../../Core/Interfaces/src/SurfelInterface.h"

#include "MySurfelList.h"
#include "MLSProjection.h"
#include "Patch.h"

/**
 * Projects the patch <tt>originalSurfelPatch</tt> onto the surface of the verification 
 * object defined by <tt>verificationSurfelList</tt>.
 *
 * @param originalSurfelPatch The patch to project.
 * @param verificationSurfelList The surface to project the patch onto.
 * @return The projected patch.
 */
SurfelCollection* MLSProjection::projectPatch(
	Patch* originalSurfelPatch, 
	MySurfelList* verificationSurfelList) 
{

	unsigned int i;
	unsigned int j;
	unsigned int surfelCount;
	SurfelInterface* surfel;

	// create data structures for the neighbourhood search on the verification surface
	unsigned int nofVerificationSurfels = verificationSurfelList->count();
	Vector3D* verificationPositions = new Vector3D[nofVerificationSurfels];
	SurfelInterface** verificationSurfels = new SurfelInterface*[nofVerificationSurfels];
	for (i = 0, surfel = verificationSurfelList->first(); i < nofVerificationSurfels; i++, surfel = verificationSurfelList->next()) {
		verificationPositions[i] = surfel->getPosition();
		verificationSurfels[i] = surfel;
	}

	// initialize the neighbourhood search objects
	NeighbourHood* neighbourHoodOnVerificationSurface = new NeighbourHood(verificationPositions, nofVerificationSurfels);

	// for efficiency we allocate some data structure already here
	// create data matrix and covariance matrix
	const unsigned int numberOfPointsForMls = 30;
	unsigned int matrixRows = numberOfPointsForMls; 
	unsigned int matrixCols = 3;
	gsl_matrix* dataMatrix = gsl_matrix_alloc(matrixRows, matrixCols);
	gsl_matrix* covarianceMatrix = gsl_matrix_alloc(matrixCols, matrixCols);
	gsl_matrix* transposedDataMatrix = gsl_matrix_alloc(matrixCols, matrixRows);
	gsl_matrix* eigenVectors = gsl_matrix_alloc(matrixCols, matrixCols);	
	gsl_vector* eigenVector = gsl_vector_alloc(matrixCols);
	gsl_vector* eigenValues = gsl_vector_alloc(matrixCols);
	gsl_eigen_symmv_workspace* workspace = gsl_eigen_symmv_alloc(matrixCols);

	// create data structure which will hold the result
	SurfelCollection* projectedPatch = new SurfelCollection();
	SurfelInterface* projectedSurfel;

	// we will count the points which do not converge well
	int nofBadConvergence = 0;

	// iterate over all the surfels of the original object and project them onto the modified surface
	unsigned int nofOriginalSurfels = originalSurfelPatch->count();
	for (surfelCount = 0, surfel = originalSurfelPatch->first(); surfelCount < nofOriginalSurfels; surfelCount++, surfel = originalSurfelPatch->next()) {

		// the following code uses an iterative approach based on a simplified MLS
		
		// get the initial estimate by choosing the nearest point.
		// set the source point to the position of the surfel to project.
		Vector3D surfelPosition = surfel->getPosition();
		neighbourHoodOnVerificationSurface->setSourcePoint(surfelPosition);
		// get the nearest neighbour
		neighbourHoodOnVerificationSurface->setNofNeighbours(1);

		// we will get the position of the first approximation
		Vector3D projectedPosition = neighbourHoodOnVerificationSurface->getNeighbourPosition(0);

		// get the MLS plane
		// for that we have to determine the normal vector using lagrange multipliers
		float dampingFactor = 0.5;
		float distance;
		float radius;
		float constD;
		int iterationCount = 0;
		Vector3D normal;
		do {
			// get the nearest neighbours and calculate the radius of the enclosing sphere
			neighbourHoodOnVerificationSurface->setSourcePoint(projectedPosition);
			neighbourHoodOnVerificationSurface->setNofNeighbours(numberOfPointsForMls);
			Vector3D points[numberOfPointsForMls];
			float distances[numberOfPointsForMls];
			radius = 0;
			for (i=0; i<numberOfPointsForMls; i++) {
				points[i] = neighbourHoodOnVerificationSurface->getNeighbourPosition(i);
				distances[i] = Vector3D::distance(projectedPosition, points[i]);
				if (distances[i]>radius) {
					radius = distances[i];
				}
			}

			// calculate the weighted sum of the ten points
			// the sigma value is chosen such that the gaussian at the border is 0.001
			float sigmaSquared = -1/log(0.001);	
			float factor = sigmaSquared*radius*radius;
			float totalWeight = 0.0f;
			Vector3D weightedSum(0,0,0);
			for (i=0; i<numberOfPointsForMls; i++) {
				float weight = exp(-(distances[i]*distances[i])/factor);
				weightedSum += points[i]*weight;
				totalWeight += weight;
			}
			weightedSum /= totalWeight;

			// iterate over all selected points and fill the data matrix with the point coordinates
			for (i=0; i<numberOfPointsForMls; i++) {
				gsl_matrix_set(dataMatrix, i, 0, (double)points[i][0]);
				gsl_matrix_set(dataMatrix, i, 1, (double)points[i][1]);
				gsl_matrix_set(dataMatrix, i, 2, (double)points[i][2]);
			}

			// subtract the weighted average from the coordinates in the matrix
			for (i=0; i<matrixCols; i++) { 
				for (j=0; j< matrixRows; j++) {
					gsl_matrix_set(dataMatrix, j, i, gsl_matrix_get(dataMatrix, j, i) - weightedSum[i]);
				}
			}

			// create the transpose matrix of the data matrix
			gsl_matrix_transpose_memcpy(transposedDataMatrix, dataMatrix);

			// calculate the product and clean up unnecessary data
			gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, transposedDataMatrix, dataMatrix, 0.0, covarianceMatrix);

			// calculate eigenvector with smallest associated eigenvalue
			// this eigenvector is equal to the normal of the plane
			gsl_eigen_symmv(covarianceMatrix, eigenValues, eigenVectors, workspace);
			gsl_eigen_symmv_sort(eigenValues, eigenVectors, GSL_EIGEN_SORT_VAL_ASC);
			double eigenValue = gsl_vector_get(eigenValues, 0);
			gsl_matrix_get_col(eigenVector, eigenVectors, 0);
			normal[0] = gsl_vector_get(eigenVector, 0);
			normal[1] = gsl_vector_get(eigenVector, 1);
			normal[2] = gsl_vector_get(eigenVector, 2);
			// the plane goes through the weighted sum of the points
			constD = -(normal[0]*weightedSum[0]+normal[1]*weightedSum[1]+normal[2]*weightedSum[2]);

			// project the surfel onto the surface we have found
			// this is the next approximation
			float lambda = (-constD-normal[0]*surfelPosition[0]-normal[1]*surfelPosition[1]-normal[2]*surfelPosition[2]);
			lambda /= (normal[0]*normal[0]+normal[1]*normal[1]+normal[2]*normal[2]);
			Vector3D newProjectedPosition = (1-dampingFactor)*projectedPosition+dampingFactor*(surfelPosition + lambda*normal);
			distance = Vector3D::distance(newProjectedPosition, projectedPosition);
			projectedPosition = newProjectedPosition;

			// iterate while the difference is larger than a factor times the radius
			// in case it does not converge (fast enough) then exit anyway
			iterationCount++;
		} while ((distance>radius*1E-3) && (iterationCount<=20));

		// detect bad convergence
		if (iterationCount>20) {
			//qDebug("MLS: Bad Convergence (%d): Original Point (%f,%f,%f). Result (%f,%f,%f)", 
			//	surfelCount, surfelPosition[0], surfelPosition[1], surfelPosition[2],
			//	projectedPosition[0], projectedPosition[1], projectedPosition[2]);
			nofBadConvergence++;
		}

#if 1
		// the following code calculates the polynomial surface

		// create two unit length spanning vectors of the plane
		float z1 = -(normal[0]*1+normal[1]*1+constD)/normal[2];
		float z2 = -(normal[0]*2+normal[1]*2+constD)/normal[2];
		Vector3D span1(1,1,z2-z1);
		span1.normalize();
		Vector3D span2 = Vector3D::crossProduct(span1, normal);
		span2.normalize();

		// prepare nearest neighbour data structure
		neighbourHoodOnVerificationSurface->setSourcePoint(projectedPosition);
		neighbourHoodOnVerificationSurface->setNofNeighbours(numberOfPointsForMls);
		Vector3D points[numberOfPointsForMls];
		float distances[numberOfPointsForMls];
		radius = 0;
		for (i=0; i<numberOfPointsForMls; i++) {
			points[i] = neighbourHoodOnVerificationSurface->getNeighbourPosition(i);
			distances[i] = Vector3D::distance(projectedPosition, points[i]);
			if (distances[i]>radius) {
				radius = distances[i];
			}
		}

		// set up least squares data structures
		gsl_matrix *X, *cov;
		gsl_vector *y, *w, *c;
		X = gsl_matrix_alloc (numberOfPointsForMls, 16);
		y = gsl_vector_alloc (numberOfPointsForMls);
		w = gsl_vector_alloc (numberOfPointsForMls);
		c = gsl_vector_alloc (16);
		cov = gsl_matrix_alloc (16, 16);
		double chisq;

		// setting up least squares equations
		float sigmaSquared = -1/log(0.001);	
		float factor = sigmaSquared*radius*radius;
		for (i=0; i<numberOfPointsForMls; i++) {
			float fi = Vector3D::dotProduct(points[i]-projectedPosition, normal);
			float ui = Vector3D::dotProduct(points[i]-projectedPosition, span1);
			float vi = Vector3D::dotProduct(points[i]-projectedPosition, span2);
			gsl_matrix_set (X,  i, 0, ui*ui*ui*vi*vi*vi);
			gsl_matrix_set (X,  i, 1, ui*ui*ui*vi*vi);
			gsl_matrix_set (X,  i, 2, ui*ui*ui*vi);
			gsl_matrix_set (X,  i, 3, ui*ui*ui);
			gsl_matrix_set (X,  i, 4, ui*ui*vi*vi*vi);
			gsl_matrix_set (X,  i, 5, ui*ui*vi*vi);
			gsl_matrix_set (X,  i, 6, ui*ui*vi);
			gsl_matrix_set (X,  i, 7, ui*ui);
			gsl_matrix_set (X,  i, 8, ui*vi*vi*vi);
			gsl_matrix_set (X,  i, 9, ui*vi*vi);
			gsl_matrix_set (X,  i,10, ui*vi);
			gsl_matrix_set (X,  i,11, ui);
			gsl_matrix_set (X,  i,12, vi*vi*vi);
			gsl_matrix_set (X,  i,13, vi*vi);
			gsl_matrix_set (X,  i,14, vi);
			gsl_matrix_set (X,  i,15, 1);
			gsl_vector_set (y, i, fi);
			float weight = exp(-(distances[i]*distances[i])/factor);
			gsl_vector_set (w, i, weight);
		}

		// solving least squares equation
	    gsl_multifit_linear_workspace * work = gsl_multifit_linear_alloc (numberOfPointsForMls, 16);
	    gsl_multifit_wlinear(X, w, y, c, cov, &chisq, work);
		float c15 = gsl_vector_get(c,15);
		gsl_matrix_free(X);
		gsl_matrix_free(cov);
		gsl_vector_free(y);
		gsl_vector_free(w);
		gsl_vector_free(c);
	    gsl_multifit_linear_free (work);

		// get the final projected position
		float height = c15;
		projectedPosition = projectedPosition+normal*height;

#endif
		// create new surfel object
		// add the projected point to the patch
		projectedSurfel = projectedPatch->addSurfel(surfel, false);

		// create color of the surfel (take nearest neighbour color)
		neighbourHoodOnVerificationSurface->setSourcePoint(projectedPosition);
		neighbourHoodOnVerificationSurface->setNofNeighbours(1);
		QColor color = verificationSurfels[neighbourHoodOnVerificationSurface->getNeighbourPositionIndex(0)]->getDiffuseColor();

		// modify the Position. this can be accomplished by using the scene even though 
		// the surfel is not really displayed
		Scene::getInstance()->setSurfelPosition(projectedSurfel, projectedPosition, false);
		Scene::getInstance()->setSurfelDiffuseColor(projectedSurfel, color.rgb(), false);
	}

	// output statistics about bad convergence
	qDebug("MLS: %d out of %d points didn't converge as desired (%2.2f Percent)", 
		nofBadConvergence, nofOriginalSurfels, nofBadConvergence*100/(float)(nofOriginalSurfels));

	// clean up
	gsl_eigen_symmv_free(workspace);
	gsl_matrix_free(dataMatrix);
	gsl_matrix_free(transposedDataMatrix);
	gsl_matrix_free(covarianceMatrix);
	gsl_matrix_free(eigenVectors);
	gsl_vector_free(eigenVector);
	gsl_vector_free(eigenValues);
	delete[] verificationPositions;
	delete neighbourHoodOnVerificationSurface;
	delete[] verificationSurfels;

	// return the result
	return projectedPatch;
}

/**
 * Returns the name of this module.
 *
 * @return The name of this module.
 */
const char* MLSProjection::getModuleName() const {
	return "MLS Projection";
}

/**
 * Returns the description of this module.
 *
 * @return The description of this module.
 */
const char* MLSProjection::getModuleDescription() const {
	return "Projects points of one patch onto the other surface using MLS.";
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
