//
// Title:   MajorityWatermark.h
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include <qobject.h>

#include "WatermarkInterface.h"

/**
 * Watermark with error correction capabilities by using
 * majority voting.
 */
class MajorityWatermark: public WatermarkInterface {

public:
	/**
	 * Constructor for the majority watermark. 
	 * The capacity can be varied, also the chip-rate (number of repetitions of 
	 * every encoded bit, for error correction) and usage-divider (portion
	 * of the frequencies used for embedding watermark).
	 * Furthermore a strength for the watermark can be specified.
	 * Note that although usage-divider has been introduced for frequency 
	 * based embedding, this watermarking class is not restricted to these
	 * frequency based approaches. Usage-divider just tells the watermarking
	 * algorithm to reserve more frequencies than required: In the end only
	 * lower frequencies are used to embed the watermark.
	 *
	 * @param capacity The length of the watermark to be used (possible values range from 1 to 160).
	 * @param watermarkStrength The strength of the watermark.
	 * @param watermarkChipRate The chip-rate of the watermark.
	 * @param watermarkUsageDivider The usage divider of the watermark.
	 * @param colorWatermarkIn Indicator whether the watermark is to be applied on color or position.
	 * @param useFastVersionIn Indicator whether the watermark has to be applied using the fast embedding algorithm.
	 */
	MajorityWatermark(unsigned int capacity, float watermarkStrength,
		unsigned int watermarkChipRate, unsigned int watermarkUsageDivider, 
		bool colorWatermarkIn, bool useFastVersionIn);

	/** 
	 * Destructor for the majority watermark.
	 *
	 * Cleans up dynamic data structures.
	 */
	~MajorityWatermark();
	
	/** 
	 * Returns the total size in bits of this watermark.
	 *
	 * @return The total size in bits of this watermark.
	 */
	const unsigned int getSizeInBits() const;

	/**
	 * Returns the size in bits of the payload of this watermark.
	 * This is the number of bits that has to be encoded eventually.
	 * 
	 * @return The size in bits of the payload of this watermark.
	 */
	const unsigned int getUsedSizeInBits() const ;

	/** 
	 * Returns the number of user-defined bits which this watermark can hold.
	 *
	 * @return The number of user-defined bits which this watermark can hold.
	 */
	const unsigned int getCapacityInBits() const ;

	/** 
	 * Returns the number of repetitions of every encoded bit. Repetition
	 * is used for error correction.
	 *
	 * @return The number of repetitions of every encoded bit.
	 */
	const unsigned int getChipRate() const ;

	/**
	 * Returns the fraction of the total size that is used to really encode
	 * any data, e.g.\ a value of 2 means that the watermark reserves a 
	 * total size 2 times as large as the payload size (this makes sense
	 * to get more frequencies and only use the lower ones for encoding. 
	 *
	 * @return	The fraction of the total size that is used to really encode
	 *			any data.
	 */
	const unsigned int getUsageDivider() const ;

	/** 
	 * Returns the strength of this watermark.
	 *
	 * @return The strength of this watermark.
	 */
	const float getStrength() const ;

	/** 
	 * Returns whether the watermark is a color watermark or a position watermark.
	 *
	 * @return Whether the watermark is a color watermark or a position watermark.
	 */
	const bool isColorWatermark() const;

	/** 
	 * Returns whether the watermark has to be embedded using a fast algorithm.
	 *
	 * @return Whether the watermark has to be embedded using a fast algorithm.
	 */
	const bool getUseFastVersion() const;

	/**
	 * Returns the watermark bit located at index <tt>index</tt>.
	 *
	 * @param index The index of the watermark bit to retrieve.
	 * @return The watermark bit located at index <tt>index</tt>.
	 */
	const bool getBit(unsigned int index) const ;

	/**
	 * Sets the watermark bit located at index <tt>index</tt>.
	 *
	 * @param index The index of the watermark bit to set.
	 * @param value The new value for the watermark bit specified by <tt>index</tt>.
	 */
	void setBit(unsigned int index, bool value);

	/** 
	 * Casts the watermark bit to a char pointer.
	 */
	operator unsigned char*();

	/** 
	 * Returns the module name of this class.
	 *
	 * @return "Majority Watermark";
	 */
	const char* getModuleName() const;

	/** 
	 * Returns the module description of this class.
	 *
	 * @return The module description of this class.
	 */
	const char* getModuleDescription() const;

protected:
	
	/**
	 * Data structure used to store the bits of the watermark.
	 */
	unsigned char* watermarkBits;

	/**
	 * Chip rate defines the number of repetitions of every bit.
	 */
	unsigned int chipRate;

	/**
	 * Usage divider specifies the fraction of the number of frqeuencies used for encoding (e.g. 2 means 1/2 of the frequencies).
	 */
	unsigned int usageDivider;

	/**
	 * Number of bits per watermark.
	 */
	unsigned int capacityInBits;

	/**
	 * The strength of the watermark.
	 */
	float strength;

	/**
	 * Is it a watermark on the color attributes?
	 */
	bool colorWatermark;

	/**
	 * Do we want to use the fast embedding algorithm.
	 */
	bool useFastVersion;
};

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
