//
// Title:   PreprocessingInterface.h
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __PREPROCESSINGINTERFACE_H_
#define __PREPROCESSINGINTERFACE_H_

#include <qobject.h>

#include "../../../Core/Interfaces/src/PluginInterface.h"
#include "../../../Core/Interfaces/src/SurfelInterface.h"
#include "Random.h"
#include "MySurfelList.h"
#include "PatchBuilderInterface.h"
#include "WatermarkInterface.h"
#include "ModuleInterface.h"

/**
 * Interface for the preprocessing step of watermarking (like object reorientation and feature extraction).
 */
class PreprocessingInterface: public ModuleInterface {

public:

	/** 
	 * Preprocesses the surfels contained in the list. This step is performed 
	 * before embedding the watermark in order to get a deterministic reorientation 
	 * of the object in space.
	 *
	 * @param surfelList The list of surfels to preprocess and reorient in space.
	 * @param scale			A vector containing the scaling of the geometry in three dimensions.
	 *						The value must make sense: It should be used to rescale a distorted 
	 *						object by using the scale of the original object.
	 */
	virtual void preprocess(MySurfelList* surfelList, Vector3D scale = Vector3D(0,0,0)) = 0;

	/**
	 * Extracts a feature of the surfels contained in the list passed to this function.
	 * A feature is a specific property of a point cloud geometry and is used to calculate
	 * a watermark which is dependent on the object to be watermarked.
	 *
	 * @param surfelList The list of surfels for which the feature has to be calculated.
	 * @return A zero-terminated feature string dependent on surfel properties.
	 */
	virtual char* extractFeature(MySurfelList* surfelList) = 0;

	/**
	 * Gets the scale of the geometry described by the list of surfels. 
	 *
	 * @param surfelList The list containing the surfels whose scale has to be determined.
	 * @return	A vector containing the scaling of the geometry in three dimensions.
	 *			The returned value must make sense: It should be used to rescale a distorted 
	 *			object by using the scale of the original object.
	 */
	virtual Vector3D getScale(MySurfelList* surfelList) = 0;

	/** 
	 * Returns the name of this module.
	 *
	 * @return The name of this module.
	 */
	virtual const char* getModuleName() const = 0;

	/**
	 * Returns the description of this module.
	 *
	 * @return The description of this module.
	 */
	virtual const char* getModuleDescription() const = 0;
};

#endif __PREPROCESSINGINTERFACE_H_

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
