//
// Title:   PrivateWatermarking.cpp
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

/**
 * @file PrivateWatermarking.cpp
 *
 * This file implements a private watermarking scheme. The original object is required for
 * watermark extraction.
 */
#include <qobject.h>
#include <qmessagebox.h>
#include <assert.h>
#include <time.h>

#include "../../../Core/Scene/src/Scene.h"
#include "../../../Core/Scene/src/Object.h"
#include "../../../Core/DataStructures/src/SurfelCollection.h"
#include "../../../Core/CoreTools/src/CoreTools.h"
#include "../../../Core/CoreTools/SelectionTool/src/SelectionTool.h"
#include "../../../Utilities/UserMessages/src/StatusBar.h"
#include "../../../Core/Interfaces/src/SurfelInterface.h"

#include "MySurfelList.h"
#include "Patch.h"
#include "PrivateWatermarking.h"
#include "WatermarkingInterface.h"
#include "MajorityWatermark.h"
#include "PatchBuilderInterface.h"
#include "RegistrationInterface.h"
#include "PCARegistration.h"
#include "PreprocessingInterface.h"
#include "PCAPreprocessing.h"
#include "ProjectionInterface.h"
#include "MLSProjection.h"
#include "PlaneProjection.h"
#include "NearestNeighbourProjection.h"
#include "PCAPatchBuilder.h"
#include "Random.h"
#include "SHA1.h"
#include "CRC32.h"
#include "LaplaceFrequencyTransformation.h"
#include "FrequencyTransformationInterface.h"
#include "CorrelationDialogImpl.h"

// the implementation of various algorithms can be changed by adapting the following typedefs
typedef MajorityWatermark WatermarkImplementation;						// class holding the watermark
typedef PCAPreprocessing PreprocessingImplementation;					// preprocessing applied to the surfels to be watermarked
typedef PCAPatchBuilder PatchBuilderImplementation;						// class used for building the patches 
typedef PCARegistration RegistrationImplementation;						// object registration at beginning of extraction
typedef MLSProjection ProjectionImplementation;						    // project a patch onto a point cloud to get another patch

// parameters for the watermarking follow:
// specifies how many attribute values per surfel we want to use for the watermarking process
// usually we will use 3 of them: x, y and z coordinate.
#define nofAttributes 3

// some typedefs for the frequency transformation:
// to replace the laplace frequency transformation by another algorithm, the following
// lines can be adapted.
typedef FrequencyTransformationInterface<nofAttributes> FTInterface;	// frequency transformation interface, convenience short version
typedef GenericVector<nofAttributes> Spectrum;							// spectrum typedef, convenience short version
typedef LaplaceFrequencyTransformation<nofAttributes> FTImplementation;	// frequency transformation implementation
typedef FTImplementation::LaplaceDataPack DataPack;						// frequency transformation implementation data pack

// temporary seed for debugging code (directly extract after embedding into a patch)
double seedx;

/**
 * Embeds a watermark into the currently selected surfels or the active object.
 * The watermark is based on a passphrase, which is passed to this function.
 * Although a selection is supported in the embedding process, the extraction procedure
 * only supports PointShop3D objects and no selection. Therefore selections have to be converted 
 * to objects before extraction.
 *
 * @param passPhrase	The secret, user defined passphrase to use for the watermark creation.
 * @param watermark		A watermark configured with the options the user specified. Note that the 
 *						watermark will be filled with content later on.
 */
void PrivateWatermarking::embedWatermark(const QString& passPhrase, WatermarkInterface* watermark) 
{
	// some data structures used for the embedding process
	StatusBar                *statusBar;
	SurfelInterface          *surfel;
	SurfelCollection         *activeSurfelCollection;
	Scene                    *scene;
	uint                     nofSurfels;
	uint					 i;
	QVector<SurfelInterface> *selection;
	SelectionTool            *selectionTool;
	statusBar    = StatusBar::getInstance();
	scene        = Scene::getInstance();

	// get the surfels of the currently active object
	activeSurfelCollection = Scene::getInstance()->getActiveObject()->getSurfelCollection();

	// reset the user information, i.e. set the progress bar to zero and show the status bar message
	statusBar->resetProgress();
	statusBar->showMessage ("Starting Embedding Watermark...", 0);

	// this will contain the list of surfels to modify 
	MySurfelList surfelList;

	// creation of the list has two different cases (selection, noselection)
	statusBar->showMessage ("Creating Surfel Lists...", 0);
	selectionTool = CoreTools::getInstance()->getSelectionTool();
	if (selectionTool->hasSelection() == false) {

		// the user hasn't selected any surfels. we will therefore use the active object
		// get the number of surfels we have to add to the list
		nofSurfels = activeSurfelCollection->getNofSurfels();
		// create list of surfels
		for (i = 0, surfel = activeSurfelCollection->getFirstSurfel(); i < nofSurfels; i++, surfel = activeSurfelCollection->getNextSurfel()) {
			surfelList.append(surfel);
		}

	} else {

		// the user has selected a portion of the surfels
		selection = selectionTool->getSelection();
		// get the number of surfels we have to add to the list
		nofSurfels = selection->size();
		// create list of surfels
		for (i = 0; i < nofSurfels - 1; i++) {
			surfel = selection->at(i);
			surfelList.append(surfel);
		}
	}

	// calculate the object extent (used as a factor for the amplitude modulation)
	float maxX = FLT_MIN;
	float maxY = FLT_MIN;
	float maxZ = FLT_MIN;
	float minX = FLT_MAX;
	float minY = FLT_MAX;
	float minZ = FLT_MAX;
	nofSurfels = activeSurfelCollection->getNofSurfels();
	for (i = 0, surfel = activeSurfelCollection->getFirstSurfel(); i < nofSurfels; i++, surfel = activeSurfelCollection->getNextSurfel()) {
		Vector3D position = surfel->getPosition();
		maxX = (position[0]>maxX)?position[0]:maxX;
		maxY = (position[1]>maxY)?position[1]:maxY;
		maxZ = (position[2]>maxZ)?position[2]:maxZ;
		minX = (position[0]<minX)?position[0]:minX;
		minY = (position[1]<minY)?position[1]:minY;
		minZ = (position[2]<minZ)?position[2]:minZ;
	}
	float objectExtent = maxX-minX;
	objectExtent = (maxY-minY>objectExtent)?maxY-minY:objectExtent;
	objectExtent = (maxZ-minZ>objectExtent)?maxZ-minZ:objectExtent;

	// create the patches
	// the current implementation uses PCA to subdivide the space recursively
	// the patches have to be created before modifying the original object
	statusBar->showMessage ("Creating Patches...", 0);
	PatchBuilderInterface* patchBuilder = new PatchBuilderImplementation(watermark);
	PatchBuilderInterface::PatchList* patchList = patchBuilder->buildPatches(&surfelList);
	delete patchBuilder; 

	// check whether the user wants to color the patches (used for debugging and education)
	switch( QMessageBox::information( NULL, "Question",	"Do you want to color the patches?\n"
								"Do not color them if attributes contain watermarks!",
								"Yes", "No", 0, 0, 1 ))	{
	case 0: 
		// Define a random number generator for the coloring of the patches
		Random randomForColor(3615);

		// now process every surfel of the object and color it
		for (Patch* colorPatch=patchList->first(); colorPatch!=0; colorPatch=patchList->next()) {
			
			// define the color of the current patch
			QColor color; 
			float red   = randomForColor.getUniform() * 256;
			float green = randomForColor.getUniform() * 256; 
			float blue  = randomForColor.getUniform() * 256;
			color.setRgb(red, green, blue);

			// apply the color to all the surfels of the current patch
			for (SurfelInterface* surfel=colorPatch->first(); surfel!=0; surfel=colorPatch->next()) {
				// modify the color
				scene->setSurfelDiffuseColor(surfel, color.rgb(), false);
			}
		}
		
		// emit the signal in order to update the screen 
		colorPatch=patchList->first();
		SurfelInterface* firstSurfel=colorPatch->first();
		QColor color = firstSurfel->getDiffuseColor();
		scene->setSurfelDiffuseColor (firstSurfel, color.rgb(), true);
	}

	// preprocessing: the current implementation uses reorientation using PCA
	statusBar->showMessage ("Preprocessing...", 0);
	PreprocessingInterface* preprocessor = new PreprocessingImplementation();
	// extract features from the point cloud (watermark has to depend on original object)
	// this task is also performed by the preprocessor
	char* feature = preprocessor->extractFeature(&surfelList); 
	// now perform the real preprocessing that can modify the original object
	// note that we first have to extract the feature before modifying the object
	switch( QMessageBox::information( NULL, "Question",	"Do you want to preprocess the object for PCA registration?\n"
								"This is only required if you want to apply PCA registration during extraction.",
								"Yes", "No", 0, 0, 1 )) {
	case 1:
		// the user wants to skip the preprocessing step
		break;
	case 0: 
		// the user wants to preprocess the surfels
		preprocessor->preprocess(&surfelList);
	}
	// we no longer need the preprocessor, so release memory
	delete preprocessor; 

	// create the watermark bit string by running SHA-1 on passphrase and feature
	statusBar->showMessage ("Generating Secure Watermark...", 0);
	SHA1 sha1;
	sha1.reset();
	sha1.update((unsigned char *)(const char *)passPhrase, passPhrase.length());
	sha1.update((unsigned char *)(const char *)feature, strlen(feature));
	sha1.final();
	unsigned char hashValue[21];
	sha1.getHash(hashValue);
	hashValue[20]=0;
	//qDebug("Hash Value at Embedding: %s", hashValue);

	// we no longer need the feature, so clean up
	free(feature);

	// propagate the hash value to the watermark
	for (unsigned int bitNumber=0; bitNumber<watermark->getCapacityInBits(); bitNumber++) {
		watermark->setBit(bitNumber, hashValue[bitNumber/8]&(1<<(bitNumber%8)));
	}

	// calculate the seed for the random number generator
	CRC32 crc32;
	double seed = crc32.calcCRC(hashValue, 20);
	seedx = seed;

	// we will output some timing information for the following code
	char statusMessage[20];
	FILE* file = fopen("log.txt", "a+");
	assert(file!=NULL);

	// apply the watermark to every patch
	uint currentPatchNumber=0;
	uint totalPatchNumber=patchList->count();
	for (Patch* patch = patchList->first(); patch!=0; patch = patchList->current()) {

		// initialize random number generator with a secret value
		delete random;
		random = new Random(seed);

		// display process status
		sprintf(statusMessage, "Patch #%04d", ++currentPatchNumber);
		statusBar->showMessage (statusMessage, 0);
		statusBar->setProgress ((int)(currentPatchNumber/(float)totalPatchNumber*100.0), true);

		// initialize timing data structures
		clock_t tv1, tv2;
		double time;
		tv1 = clock();

		// process current patch
		embedWatermarkIntoPatch(patch, watermark, objectExtent);

		// calculate the time elapsed during the processing of the patch 
		// and output the information to the file and flush it
		tv2 = clock();
		time = (tv2 - tv1)/((double)CLOCKS_PER_SEC);
		//fprintf(file, "Timing: %f s for patch with %u points and %u eigenvalues.\n", time, patch->count(), watermark->getUsedSizeInBits());
		fprintf(file, "%f, %u, %u\n", time, patch->count(), watermark->getUsedSizeInBits());
		fflush(file);

		patchList->remove(patch);
		delete patch;
	}

	// clean up, i.e. close timing information file and delete the list of patches
	fclose(file);
	delete patchList;
	delete watermark;

	// we emit the signal in order to update the screen
	statusBar->showMessage ("Updating Screen...", 0);
	surfel = activeSurfelCollection->getFirstSurfel();
	scene->setSurfelPosition (surfel, surfel->getPosition(), true);

	// show message dialog box to inform the user of task completion
	QMessageBox::information( NULL, "Information", "Watermark has been embedded.\nYou can now save your watermarked object." );

	// reset status bar message and set progress bar to 100%
	statusBar->setProgress (100, true);
	statusBar->showMessage ("", 1);
}

/**
 * Embeds the watermark into the patch.
 *
 * @param patch The patch to embed the watermark into.
 * @param watermark The watermark to embed into the patch.
 * @param objectExtent The extent of the bounding box of the object.
 */
void PrivateWatermarking::embedWatermarkIntoPatch(Patch* patch, WatermarkInterface* watermark, float objectExtent) 
{

	uint i;
	StatusBar* statusBar = StatusBar::getInstance();

	// init transformation data structure
	FTInterface* frequencyTransformation = new	FTImplementation(watermark);	// the algortihm object for transformation
	FTInterface::SpectralList* spectralList;									// will contain the result of the transformation
	DataPack* dataPack;															// contains additional information allowing backward transformation
	
	// process forward transformation
	statusBar->showMessage ("Forward Transformation...", 0);
	frequencyTransformation->forwardTransformation(patch, &spectralList, (FTInterface::DataPack**)&dataPack);
	
	// embed watermark into the frequencies
	statusBar->showMessage ("Embedding...", 0);
	unsigned int bitPosition = 0;									// which bit do we have to embed into the current frequency?
	int spectrumPosition = 0;										// which spectral coefficient are we processing now?
	float watermarkStrength = watermark->getStrength()*objectExtent;// strength of frequency modulation

	qDebug("Extent %f, factor %f", objectExtent, watermarkStrength);
	// embed data into all the spectral values which have been computed
	// note that more than watermark->getCapacity() bits can be embedded if the chip-rate is set 
	// to a higher value than 1. in that case we will simply wrap around and restart with the 
	// embedding of the first bit after having embedded the last one.
	for (Spectrum* spectralCoefficient = spectralList->first(); spectralCoefficient!=0; spectralCoefficient = spectralList->next()) {

		// calculate factor based on  generated random value (-1 or 1)
		float factor;
		if (random->getUniform()<0.5) {
			// p = -1;
			// float factor = watermarkStrength*p;
			// float factor = (watermarkStrength+1.0-exp(-spectrumPosition))*p;
			factor = -watermarkStrength;
		} else {
			// p = 1;
			// factor = watermarkStrength*p;
			// factor = (watermarkStrength+1.0-exp(-spectrumPosition))*p;
			factor = +watermarkStrength;
		} 

		// modify spectrum depending on watermark bit.
		if (watermark->getBit(bitPosition)) {
			// watermark bit is set to 1
			// embed the same bit into all the three attributes
			for (i=0; i<nofAttributes; i++) {
				(*spectralCoefficient)[i] += factor;
			}
		} else {
			// watermark bit is set to 0
			// embed the same bit into all the three attributes
			for (i=0; i<nofAttributes; i++) {
				(*spectralCoefficient)[i] -= factor;
			}
		}		

		// update the position of the next spectral coefficient
		spectrumPosition++;

		// update position of next watermark bit to embed
		bitPosition++;
		if (bitPosition>=watermark->getCapacityInBits()) {
			// automatically wrap around and embed the first bit again
			bitPosition=0;
		}
	}

	
	// xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
	// this code is used to immediately read the watermark after embedding it into the patch
	/*
	SurfelCollection* collection = new SurfelCollection();
	Patch* savePatch = new Patch();
	SurfelInterface* saveSurfel;
	SurfelInterface* surfel;

	for (i = 0, surfel = patch->first(); i < patch->count(); i++, surfel = patch->next()) {

		// create new surfel object
		saveSurfel = collection->addSurfel(surfel, false);

		// add the projected point to the patch
		savePatch->append(saveSurfel);
	}
	*/
	// xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
	

	// process backward/inverse transformation
	statusBar->showMessage ("Backward Transformation...", 0);
	// note that the spectral list and the data pack are input parameters this time
	frequencyTransformation->backwardTransformation(patch, spectralList, dataPack);

	// clean up memory
	delete spectralList;
	delete dataPack;
	
	// xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
	// this code is used to immediately read the watermark after embedding it into the patch
	/*
	// this will be our threshold value
	float currentThreshold = 0.75f;

	// extract watermark from every patch

	// initialize random number generator with a secret value.
	// note that this is performed at the same point of time as during the embedding to ensure same random sequences.
	delete random;
	random = new Random(seedx);

	// stores the correlation between the original watermark and the extracted watermark.
	// note that this has to be done on the heap
	float currentCorrelation;

	// now extract the watermark from this patch
	currentCorrelation = extractWatermarkFromPatch(savePatch, patch, watermark);

	savePatch->clear();
	collection->clear();
	delete collection;
	delete savePatch;
	qDebug("Extracted from patch");

	// show correlation in numeric format
	static int correlationCount = 0;
	correlationCount++;
	static float correlation = 0.0f;
	static float maxCorrelation = 0.0f;
	correlation += currentCorrelation;
	if (currentCorrelation>maxCorrelation) {
		maxCorrelation=currentCorrelation;
	}
	if (correlationCount == 20) {
		char message[100];
		sprintf(message, "Average: %f / Maximum: %f",correlation/correlationCount, maxCorrelation);
		QMessageBox::information( NULL, "Correlation Information", message);
		correlationCount = 0;
		correlation = 0.0f;
		maxCorrelation = 0.0f;
	}
	*/
	// xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
	
	// free memory
	delete frequencyTransformation;
}

/**
 * Extracts the private watermark by using the test object and the original object. Note that
 * both geometries have to be available as objects. No selection is supported during extraction.
 * That's the place where our private watermarking algorithm is significantly different from the 
 * public one (here we need two objects: the original one and the object to be verified).
 * Since embedding is dependent on a passphrase, the same passphrase has to be used for extraction.
 *
 * @param passPhrase			The passphrase to be used for extraction.
 * @param originalWatermark		A watermark configured with the options the user specified. Note that the 
 *								watermark will be filled with content later on.
 */
void PrivateWatermarking::extractWatermark(const QString& passPhrase, WatermarkInterface* originalWatermark) {

	uint					 i;
	StatusBar                *statusBar;
	SurfelInterface          *surfel;
	Scene                    *scene;

	Object					 *originalObject;
	Object					 *verificationObject;			
	SurfelCollection         *originalSurfelCollection;
	SurfelCollection         *verificationSurfelCollection;
	uint                     nofOriginalSurfels;
	uint					 nofVerificationSurfels;

	// show message in status bar
	statusBar = StatusBar::getInstance();
	statusBar->resetProgress();
	statusBar->showMessage ("Extracting Watermark...", 0);

	// get instance of the scene for later usage
	scene = Scene::getInstance();

	// make sure we have loaded at least two different objects
	// if this is not the case, the extraction procedure will exit
	if (scene->getNofObjects()!=2) {
		QMessageBox::information( NULL, "Error", "Extraction Failed.\nTwo objects in scene required." );
		return;
	}

	// get two objects from the scene
	Object* firstObject = scene->getFirstObject();
	firstObject->getName();
	Object* secondObject = scene->getNextObject();

	// select the first object and ask the user if this is the original object
	// this is performed by scaling the first object and asking the user
	// if it is the original one
	statusBar->showMessage ("Object Assignment...", 0);
	Vector3D firstObjectScale = firstObject->getScale();
	QMessageBox::information( NULL, "Information", "I need your assistance.\n\nOne of the object will be zooming.\nTell me if it's the test object.\n\nClick on OK to proceed..." );
	for (i=0; i<10; i++) {
		firstObject->setScale(firstObjectScale*(1+i/10.0f), true);
	}
	firstObject->setScale(firstObjectScale, true);

	// create the message string displayed to the user
	char message[1024];
	sprintf(message,
        "So, was the original object moving or was\n"
        "it the object to be verified/tested?\n\n"
		"Moving object name: %s\n"
		"Other object name: %s",
		firstObject->getName().latin1(),
		secondObject->getName().latin1());

	// now ask the user and assign the objects based on the response
	switch( QMessageBox::information( NULL, "Question",	message,
								"Test/Verification Object", "Original Object", 0, 0, 1 ))
	{
		case 1: 
			// the original object was moving
			originalObject = firstObject;
			verificationObject = secondObject;
			break;
		case 0:
			// the verification object was moving
			originalObject = secondObject;
			verificationObject = firstObject;
			break;
	}

	// get the collection of surfels of both objects
	originalSurfelCollection = originalObject->getSurfelCollection();
	verificationSurfelCollection = verificationObject->getSurfelCollection();

	// create list of surfels
	// note that in this watermarking scheme we have to create two sets of surfels
	// one set contains the original object, whereas the second one contains to object to be verified
	statusBar->showMessage ("Creating Surfel List...", 0);
	
	// first create the surfel list of the original object
	MySurfelList originalSurfelList;
	nofOriginalSurfels = originalSurfelCollection->getNofSurfels();
	for ( i = 0, surfel = originalSurfelCollection->getFirstSurfel(); 
		  i < nofOriginalSurfels; 
		  i++, surfel = originalSurfelCollection->getNextSurfel()) 
	{
		originalSurfelList.append(surfel);
	}
	
	// second create the surfel list of the object to be verified
	MySurfelList verificationSurfelList;
	nofVerificationSurfels = verificationSurfelCollection->getNofSurfels();
	for ( i = 0, surfel = verificationSurfelCollection->getFirstSurfel(); 
		  i < nofVerificationSurfels; 
		  i++, surfel = verificationSurfelCollection->getNextSurfel()) 
	{
		verificationSurfelList.append(surfel);
	}

	// get the patches here using PCA space subdivision.
	// note that the patch generator is deterministic although it contains a random number generator.
	// this is because the PRNG is always initialized to the same value. because we use the original
	// object again, we will get the same patches we got when embedding the watermark
	// patches have to be generated before modifying and preprocessing the original object
	statusBar->showMessage ("Creating Patches...", 0);
	PatchBuilderInterface* patchBuilder = new PatchBuilderImplementation(originalWatermark);
	PatchBuilderInterface::PatchList* originalPatchList = patchBuilder->buildPatches(&originalSurfelList);
	delete patchBuilder; 

	// check whether the user want to color the patches (used for debugging and education)
	switch( QMessageBox::information( NULL, "Question",	"Do you want to color the patches?\n"
								"Do not color them if attributes contain watermarks!",
								"Yes", "No", 0, 0, 1 ))	{
	case 0: 
		// Define a random number generator for the coloring of the patches
		Random randomForColor(3615);

		// now process every surfel of the object and color it
		for (Patch* colorPatch=originalPatchList->first(); colorPatch!=0; colorPatch=originalPatchList->next()) {
			
			// define the color of the current patch
			QColor color; 
			float red   = randomForColor.getUniform() * 256;
			float green = randomForColor.getUniform() * 256; 
			float blue  = randomForColor.getUniform() * 256;
			color.setRgb(red, green, blue);

			// apply the color to all the surfels of the current patch
			for (SurfelInterface* surfel=colorPatch->first(); surfel!=0; surfel=colorPatch->next()) {
				// modify the color
				scene->setSurfelDiffuseColor(surfel, color.rgb(), false);
			}
		}
		
		// emit the signal in order to update the screen 
		colorPatch=originalPatchList->first();
		SurfelInterface* firstSurfel=colorPatch->first();
		QColor color = firstSurfel->getDiffuseColor();
		scene->setSurfelDiffuseColor (firstSurfel, color.rgb(), true);
		break;
	}

	// the generated secret seed must depend on some features of the original object,
	// i.e. eigenvalues of PCA for instance. this prevents some forgery attacks.
	// the same feature has been extracted while embedding the watermark
	// note that the feature has to be extracted before any modification on the 
	// original object is performed (e.g. registration)
	statusBar->showMessage ("Extracting Feature...", 0);
	PreprocessingInterface* preprocessor = new PreprocessingImplementation();
	char* feature = preprocessor->extractFeature(&originalSurfelList); 
	// show feature dialog box
	// QMessageBox::information(NULL, "Extracted Feature", feature);
	delete preprocessor; 

	// create the watermark bit string
	// process passphrase (first SHA-1, then CRC to seed the RNG)
	SHA1 sha1;
	sha1.reset();
	sha1.update((unsigned char *)(const char *)passPhrase, passPhrase.length());
	sha1.update((unsigned char *)(const char *)feature, strlen(feature));
	sha1.final();
	unsigned char hashValue[21];
	sha1.getHash(hashValue);
	hashValue[20]=0;
	// qDebug("Hash Value at Extraction: %s", hashValue);
	free(feature);

	// note that this watermark has to be the same type as the watermark embedded into
	// the object because it will be passed to some of the following routines to 
	// initialize data correctly (watermark capacity determines for instance the size 
	// of the patches).
	// final watermark generation step: as you can see it depends
	// on the feature of the original object and the passphrase (both are needed).
	for (unsigned int bitNumber=0; bitNumber<originalWatermark->getCapacityInBits(); bitNumber++) {
		originalWatermark->setBit(bitNumber, hashValue[bitNumber/8]&(1<<(bitNumber%8)));		
	}

	// we want to make the random number generator dependent on the previously calculated
	// secret hash value. unfortunately 160 bit is too long, so we use a simple CRC32 
	// routine to map the watermark value to 32 bit which corresponds to a double value.
	// this can be used to seed the PRNG.
	CRC32 crc32;
	double seed = crc32.calcCRC(hashValue, 20);
	
	// registration: the current implementation uses PCA reorientation.
	// registration already performs PCA preprocessing, so this step is no longer needed.
	switch( QMessageBox::information( NULL, "Question",	"Do you want to register both objects?\n"
								"This step is recommended if your objects are not well aligned.",
								"Yes", "No", 0, 0, 1 ))
	{
	case 1:
		// the user wants to skip the registration process
		break;
	case 0: 
		// the user wants to register both objects
		statusBar->showMessage ("Registering Objects...", 0);
		RegistrationInterface* registrator = new RegistrationImplementation();
		registrator->registerObjects(&originalSurfelList, &verificationSurfelList);
		delete registrator;
		break;
	}

	// lists in which we will store the results to display later on the screen in a graph.
	QList<float> correlationList;	// list containing the correlation values of the various patches.
	QList<float> thresholdList;		// list that is supposed to contain the threshold values.
									// since we do not have any calculation for threshold values in the private watermarking scheme,
									// we will put some meaningful values.
									// this allows us to use the same dialog box as in the public watermarking scheme.

	// this will be our threshold value
	float* currentThreshold = new float(0.75f);

	// we will output some status information
	char statusMessage[20];

	// apply the watermark to every patch
	uint currentPatchNumber=0;
	uint totalPatchNumber=originalPatchList->count();

	// ask user if he wants to color the patches according to the success of extraction
	bool colorPatchesAccordingToSuccess;
	switch( QMessageBox::information( NULL, "Question",	"Do you want to color the patches according to the success of extraction?",
								"Yes", "No", 0, 0, 1 ))
	{
		case 0: 
			colorPatchesAccordingToSuccess = true;
			break;
		case 1:
			colorPatchesAccordingToSuccess = false;
			break;
	}

	// extract watermark from every patch
	for (	Patch* patch = originalPatchList->first(); 
			patch!=0; 
			patch = originalPatchList->current()) 
	{

		// initialize random number generator with a secret value.
		// note that this is performed at the same point of time as during the embedding to ensure same random sequences.
		delete random;
		random = new Random(seed);

		// stores the correlation between the original watermark and the extracted watermark.
		// note that this has to be done on the heap
		float* currentCorrelation = new float;

		// display process status
		sprintf(statusMessage, "Patch #%04d", ++currentPatchNumber);
		statusBar->showMessage (statusMessage, 0);
		statusBar->setProgress ((int)(currentPatchNumber/(float)totalPatchNumber*100.0), true);

		// now extract the watermark from this patch
		*currentCorrelation = extractWatermarkFromPatch(patch, &verificationSurfelList, originalWatermark);

		// check if the user wishes to color the patches according to extraction success
		if (colorPatchesAccordingToSuccess) {
			// define the success color of the current patch
			QColor color; 
			int red   = (*currentThreshold-*currentCorrelation)/(*currentThreshold)*127+128;
			int green = (*currentCorrelation-*currentThreshold)/(1-*currentThreshold)*127+128; 
			if (*currentCorrelation-*currentThreshold<=0) {
				green=0;
			} else {
				red=0;
			}
			int blue  = 0;
			color.setRgb(red, green, blue);

			// apply the color to all the surfels of the current patch
			for (surfel=patch->first(); surfel!=0; surfel=patch->next()) {
				// modify the color
				scene->setSurfelDiffuseColor(surfel, color.rgb(), false);
			}
		}
		originalPatchList->remove(patch);
		delete patch;

		// process the resulting values: output debug information and add the value to the list.
		// qDebug("Correlation: %f\r\n", *currentCorrelation);
		correlationList.append(currentCorrelation);

		// in this watermarking scheme there is no calculation for threshold values. 
		// we will simply put some meaningful values.
		thresholdList.append(currentThreshold);
	}
	delete originalPatchList;
	delete originalWatermark;

	// reset the status bar and progress bar
	statusBar->setProgress (100, true);
	statusBar->showMessage ("", 1);

	// print message for the user
	switch (QMessageBox::information( NULL, "Information", 
			"Watermark has been extracted.\nResults can be seen in a graph.\n",
			"Show graph", "Skip graph", 0, 0, 1))
	{
	case 1:
		// do not show any information
		break;
	case 0:
		// print the correlation dialog box
		CorrelationDialogImpl correlationDialog(&correlationList, &thresholdList, 0, "Watermark Correlation for every Patch", true);
		correlationDialog.show();
		correlationDialog.raise();
		break;
	}

	// should clean up lists
	correlationList.clear();
	thresholdList.clear();
}

/**
 * Extracts watermark from a patch and a verification object.
 *
 * @param originalPatch				The patch to be processed (on the original object).
 * @param verificationSurfelList	The list containing all the points from the object to be verified.
 * @param originalWatermark			The watermark that is supposed to result from the extraction. It will be
 *									compared with the watermark that is extracted and a correlation value will
 *									be calculated.
 * @return							The correlation value between the specified watermark and the extracted one.
 */
float PrivateWatermarking::extractWatermarkFromPatch(
		Patch* originalPatch,					// patch to process
		MySurfelList* verificationSurfelList,	// project onto this
		WatermarkInterface* originalWatermark)	// compare extracted watermark with this
{

	uint i;
	StatusBar* statusBar = StatusBar::getInstance();

	// init transformation data structure for the points of the original object (patch)
	FTInterface* frequencyTransformation = new	FTImplementation(originalWatermark);
	FTInterface::SpectralList* originalSpectralList; // this will contain the spectral components of the transformed object
	DataPack* originalDataPack;						 // this will contain some temporary data used for backward transformation
	
	// process forward transformation, result will be contained in originalSpectralList and originalDataPack.
	statusBar->showMessage ("Forward Transformation (Original)...", 0);
	frequencyTransformation->forwardTransformation(
		originalPatch, 
		&originalSpectralList, 
		(FTInterface::DataPack**)&originalDataPack);

	// project points of the original patch onto the object to be verified.
	// this gives us surfels of a patch on the verification object resampled at the original rate.
	statusBar->showMessage ("Projecting Patch...", 0);
	Patch* resampledVerificationPatch = new Patch();
	ProjectionInterface* projection = new ProjectionImplementation();
	SurfelCollection* collection = projection->projectPatch(originalPatch, verificationSurfelList);
	for (SurfelInterface* surfel = collection->getFirstSurfel(); surfel!=0; surfel=collection->getNextSurfel()) {
		resampledVerificationPatch->append(surfel);
	}
	delete projection;
	
	// we can now transform the points on the verification patch we calculated in the previous step
	// process forward transformation, result will be contained in verificationSpectralList and verificationDataPack.
	FTInterface::SpectralList* verificationSpectralList;
	statusBar->showMessage ("Forward Transformation (Verification)...", 0);
	frequencyTransformation->forwardTransformation(
		resampledVerificationPatch,
		originalDataPack,
		&verificationSpectralList);

	// for every bit of the watermark we have to count the number of
	// times we extract it as zero or one. 
	float* sumArray = new float[originalWatermark->getCapacityInBits()];

	// intialize with zero values
	for (i=0; i<originalWatermark->getCapacityInBits(); i++) {
		sumArray[i] = 0.0f;
	}

	// we can now calculate the watermark, that is extract it by simply comparing the frequencies.
	// however we will first not directly extract it, but count the occurences of ones and zeroes.
	statusBar->showMessage ("Extracting Watermark...", 0);
	unsigned int bitPosition = 0;				// bit position inside watermark
	Spectrum* originalSpectralCoefficient;		// one spectral coefficient of the original object
	Spectrum* verificationSpectralCoefficient;	// one spectral coefficient of the verification object
	for (originalSpectralCoefficient = originalSpectralList->first(),
		 verificationSpectralCoefficient = verificationSpectralList->first();
		 originalSpectralCoefficient!=0; 
		 originalSpectralCoefficient = originalSpectralList->next(),
		 verificationSpectralCoefficient = verificationSpectralList->next()) 
	{
		int p;
		if (random->getUniform()<0.5) {
			p = -1;
		} else {
			p = 1;
		}

		// verbose output (comment out in final version)
		/*
		bool bitS = ((p*(*verificationSpectralCoefficient)[0])>(p*(*originalSpectralCoefficient)[0]));
		bool bitT = ((p*(*verificationSpectralCoefficient)[1])>(p*(*originalSpectralCoefficient)[1]));
		bool bitU = ((p*(*verificationSpectralCoefficient)[2])>(p*(*originalSpectralCoefficient)[2]));
		if (originalWatermark->getBit(bitPosition)==bitS) { 
			qDebug("Bit %02d:%01d >-SUCCESS s: %f %f", bitPosition, originalWatermark->getBit(bitPosition), (*originalSpectralCoefficient)[0], (*verificationSpectralCoefficient)[0]);
		} else {
			qDebug("Bit %02d:%01d >-FAILED  s: %f %f", bitPosition, originalWatermark->getBit(bitPosition), (*originalSpectralCoefficient)[0], (*verificationSpectralCoefficient)[0]);
		}
		if (originalWatermark->getBit(bitPosition)==bitT) { 
			qDebug("         | SUCCESS t: %f %f", (*originalSpectralCoefficient)[1], (*verificationSpectralCoefficient)[1]);
		} else {
			qDebug("         | FAILED  t: %f %f", (*originalSpectralCoefficient)[1], (*verificationSpectralCoefficient)[1]);
		} 
		if (originalWatermark->getBit(bitPosition)==bitU) { 
			qDebug("         | SUCCESS u: %f %f", (*originalSpectralCoefficient)[2], (*verificationSpectralCoefficient)[2]);
		} else {
			qDebug("         | FAILED  u: %f %f", (*originalSpectralCoefficient)[2], (*verificationSpectralCoefficient)[2]);
		}
		*/

		for (i=0; i<nofAttributes; i++) {
			sumArray[bitPosition] += p*((*verificationSpectralCoefficient)[i] - (*originalSpectralCoefficient)[i]);
		}

		// update position of next watermark bit to embed
		bitPosition++;
		if (bitPosition>=originalWatermark->getCapacityInBits()) {
			// automatically wrap around and restart with first bit
			bitPosition=0;
		}
	}

	// now calculate the correlation value
	// initial correlation is set to zero
	statusBar->showMessage ("Calculating Correlation...", 0);
	int count = 0;
	float correlation = 0.0f;					
	for (i=0; i<originalWatermark->getCapacityInBits(); i++) {
		// we set the extracted watermark bit to one in case we encountered more ones than zeroes.
        // bool extractedWatermarkBit = (countOne[i]>countZero[i]);
        bool extractedWatermarkBit = (sumArray[i]>0);
		bool originalWatermarkBit = (originalWatermark->getBit(i)==1);

		// check if the bit corresponds to the bit value of the original watermark
		if (extractedWatermarkBit == originalWatermarkBit) {
			correlation += 1.0f;
		}
		count++;		
	}	

	// final step of correlation calculation (rescaling)
	correlation *= 1.0f/float(count);

	// clean up
	resampledVerificationPatch->clear();
	collection->clear();
	delete collection;
	delete[] sumArray;
	delete frequencyTransformation;
	delete resampledVerificationPatch;
	delete originalDataPack;
	delete originalSpectralList;
	delete verificationSpectralList;

	// return the correlation value
	return correlation;
}

/**
 * Returns the name of this module.
 *
 * @return "Private Watermark"
 */
const char* PrivateWatermarking::getModuleName() const {
	return "Private Watermark";
}
  
/**
 * Returns the description of this module.
 *
 * @return The description of this module.
 */
const char* PrivateWatermarking::getModuleDescription() const {
	return	"This algorithm implements a private watermark.\n"
			"Verification requires the original object and a\n"
			"test object. Embedding is performed by modifying\n"
			"the spectral coefficients computed by solving an\n"
			"eigenproblem on the laplacian matrix of the points.";
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
