//
// Title:   PublicWatermarking.cpp
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

/**
 * @file PublicWatermarking.cpp
 *
 * Implements a public watermarking scheme where the original object is no longer 
 * required when one wants to extract the watermark. This code has been implemented
 * for testing purposes. Experiments have shown that the approach chosen in this
 * file is currently not suited for real world applications because of the 
 * verly long runtime. The extraction is based on correlation calculation which 
 * only performs well for large watermarks with a few thousands of bits. 
 * To embed so many bits the frequency transformation requires a huge amount of 
 * time.
 * For the reasons mentioned above this code has not been completed and represents
 * more a work in progress.
 */

#include <qobject.h>
#include <qmessagebox.h>
#include <assert.h>

#include "../../../Core/Scene/src/Scene.h"
#include "../../../Core/Scene/src/Object.h"
#include "../../../Core/DataStructures/src/SurfelCollection.h"
#include "../../../Core/CoreTools/src/CoreTools.h"
#include "../../../Core/CoreTools/SelectionTool/src/SelectionTool.h"
#include "../../../Utilities/UserMessages/src/StatusBar.h"
#include "../../../Core/Interfaces/src/SurfelInterface.h"

#include "Patch.h"
#include "MySurfelList.h"
#include "PublicWatermarking.h"
#include "WatermarkingInterface.h"
#include "MajorityWatermark.h"
#include "PatchBuilderInterface.h"
#include "PreprocessingInterface.h"
#include "PCAPreprocessing.h"
#include "PCAPatchBuilder.h"
#include "Random.h"
#include "CorrelationDialogImpl.h"
#include "SHA1.h"
#include "CRC32.h"
#include "LaplaceFrequencyTransformation.h"
#include "FrequencyTransformationInterface.h"

// debug function prototype. this is used for testing the correlation based watermarking approach
debug();

// the implementation of various algorithms can be changed by adapting the following typedefs
typedef MajorityWatermark WatermarkImplementation;						// class holding the watermark
typedef PCAPreprocessing PreprocessingImplementation;					// preprocessing applied to the surfels to be watermarked
typedef PCAPatchBuilder PatchBuilderImplementation;						// class used for building the patches 

// parameters for the watermarking follow:
// specifies how many attribute values per surfel we want to use for the watermarking process
// usually we will use 3 of them: x, y and z coordinate.
#define nofAttributes 3

// some typedefs for the frequency transformation:
// to replace the laplace frequency transformation by another algorithm, the following
// lines can be adapted.
typedef FrequencyTransformationInterface<nofAttributes> FTInterface;	// frequency transformation interface, convenience short version
typedef GenericVector<nofAttributes> Spectrum;							// spectrum typedef, convenience short version
typedef LaplaceFrequencyTransformation<nofAttributes> FTImplementation;	// frequency transformation implementation


/**
 * Embeds a watermark into the currently selected surfels or the active object.
 * The watermark is based on a passphrase, which is passed to this function.
 *
 * @param passPhrase	The secret, user defined passphrase to use for the watermark creation.
 * @param watermark		A watermark configured with the options the user specified. Note that the 
 *						watermark will be filled with content later on.
*/
void PublicWatermarking::embedWatermark(const QString& passPhrase, WatermarkInterface* watermark) {

	// Please uncomment the following two lines if you want to test the correlation based 
	// watermarking approach instead of really embedding a watermark into the object.
	 debug();
	 return;

	// declare and define some data structures
	StatusBar                *statusBar;
	SurfelInterface          *surfel;
	SurfelCollection         *activeSurfelCollection;
	Scene                    *scene;
	uint                     nofSurfels;
	uint					 i;
	QVector<SurfelInterface> *selection;
	SelectionTool            *selectionTool;
	statusBar    = StatusBar::getInstance();
	scene        = Scene::getInstance();

	// inform the user of the start of the embedding process
	statusBar->resetProgress();
	statusBar->showMessage ("Starting Embedding Watermark...", 0);

	// get the surfels of the active object
	activeSurfelCollection = Scene::getInstance()->getActiveObject()->getSurfelCollection();

	// create list of the surfels to embed a watermark into
	MySurfelList surfelList;
	statusBar->showMessage ("Creating Surfel List...", 0);

	// two different cases (selection, noselection)
	selectionTool = CoreTools::getInstance()->getSelectionTool();
	if (selectionTool->hasSelection() == false) {

		// no selection defined by the user, so use the active object
		// get the number of surfels contained in the active object
		nofSurfels = activeSurfelCollection->getNofSurfels();

		// create list of surfels
		for (i = 0, surfel = activeSurfelCollection->getFirstSurfel(); i < nofSurfels; i++, surfel = activeSurfelCollection->getNextSurfel()) {
			surfelList.append(surfel);
		}
	} else {

		// user has selected some surfels
		selection = selectionTool->getSelection();
		nofSurfels = selection->size();

		// create list of surfels
		for (i = 0; i < nofSurfels - 1; i++) {
			surfel = selection->at(i);
			surfelList.append(surfel);
		}
	}

	// preprocessing: the current implementation uses reorientation using PCA
	statusBar->showMessage ("Preprocessing...", 0);
	PreprocessingInterface* preprocessor = new PreprocessingImplementation();
	// extract features from the point cloud (watermark has to depend on original object)
	// this task is also performed by the preprocessor
	char* feature = preprocessor->extractFeature(&surfelList); 
	// now perform the real preprocessing that can modify the original object
	// note that we first have to extract the feature before modifying the object
	preprocessor->preprocess(&surfelList);
	// we no longer need the preprocessor, so release memory
	delete preprocessor; 

	// create the watermark bit string by running SHA-1 on passphrase and feature
	statusBar->showMessage ("Generating Secure Watermark...", 0);
	SHA1 sha1;
	sha1.reset();
	sha1.update((unsigned char *)(const char *)passPhrase, passPhrase.length());
	sha1.update((unsigned char *)(const char *)feature, strlen(feature));
	sha1.final();
	unsigned char hashValue[20];
	sha1.getHash(hashValue);

	// we have to store the feature (it will remain a secret of the owner of the object
	// the feature is needed again during extraction
	FILE * file = fopen("feature.txt", "w+");
	fputs(feature, file);
	fflush(file);
	fclose(file);

	// inform the user that he has to store the feature file
	// show message dialog box to inform the user of task completion
	QMessageBox::information(NULL, "Information", "Feature has been written to 'feature.txt'.\nThis file is needed for extraction, so please\nstore it to a safe place.");

	// we no longer need the feature, so clean up
	delete[] feature;

	// create the watermark object already now because some other algorithm may depend on
	// some properties of the watermark (size...)
	// the current implementation uses a majority watermark
	for (unsigned int bitNumber=0; bitNumber<watermark->getCapacityInBits(); bitNumber++) {
		watermark->setBit(bitNumber, hashValue[bitNumber/8]&(1<<(bitNumber%8)));
	}

	// calculate the seed for the random number generator
	CRC32 crc32;
	double seed = crc32.calcCRC(hashValue, 20);

	// create the patches
	// the current implementation uses PCA to subdivide the space recursively
	statusBar->showMessage ("Creating Patches...", 0);
	PatchBuilderInterface* patchBuilder = new PatchBuilderImplementation(watermark);
	PatchBuilderInterface::PatchList* patchList = patchBuilder->buildPatches(&surfelList);
	delete patchBuilder; 
	
	// we will output some information for the following code
	char statusMessage[20];

	// apply the watermark to every patch
	uint currentPatchNumber=0;
	uint totalPatchNumber=patchList->count();
	for (Patch* patch = patchList->first(); patch!=0; patch = patchList->next()) {

		// initialize random number generator with a secret value
		delete random;
		random = new Random(seed);

		// display process status
		sprintf(statusMessage, "Patch #%04d", ++currentPatchNumber);
		statusBar->showMessage (statusMessage, 0);
		statusBar->setProgress ((int)(currentPatchNumber/(float)totalPatchNumber*100.0), true);

		// process current patch
		embedWatermarkIntoPatch(patch, watermark);
	}
	
	// clean up, i.e. delete the list of patches
	delete patchList;

	// we emit the signal in order to update the screen
	statusBar->showMessage ("Updating Screen...", 0);
	surfel = activeSurfelCollection->getFirstSurfel();
	scene->setSurfelPosition (surfel, surfel->getPosition(), true);

	// print message for the user
	QMessageBox::information( NULL, "Information", "Watermark has been embedded.\nYou can now save your watermarked object." );

	// reset status bar message and set progress bar to 100%
	statusBar->setProgress (100, true);
	statusBar->showMessage ("", 1);
}

/**
 * Embeds the watermark into the patch.
 *
 * @param patch		The patch to embed the watermark into.
 * @param watermark The watermark to embed into the patch. This object also contains the strength 
 *					of the watermark (amplitude of spectral modifications while embedding). 
 */
void PublicWatermarking::embedWatermarkIntoPatch(Patch* patch, WatermarkInterface* watermark) {

	uint i;
	StatusBar* statusBar = StatusBar::getInstance();

	// init transformation data structure
	FTInterface* frequencyTransformation = new	FTImplementation(watermark);	// the algortihm object for transformation
	FTInterface::SpectralList* spectralList;									// will contain the result of the transformation
	FTInterface::DataPack* dataPack;											// contains additional information allowing backward transformation
	
	// process forward transformation
	statusBar->showMessage ("Forward Transformation...", 0);
	frequencyTransformation->forwardTransformation(patch, &spectralList, &dataPack);
	
	// embed watermark into the frequencies
	statusBar->showMessage ("Embedding...", 0);
	float watermarkStrength = watermark->getStrength();
	int bitPosition = 0;
	for (Spectrum* spectralCoefficient = spectralList->first(); spectralCoefficient!=0; spectralCoefficient = spectralList->next()) {

		float factor;
		if (watermark->getBit(bitPosition)) {
			factor = watermarkStrength*random->getGaussian();
			for (i=0; i<nofAttributes; i++) {
				(*spectralCoefficient)[i] += fabs((*spectralCoefficient)[i])*factor;
			}
		} else {
			factor = -watermarkStrength*random->getGaussian();
			for (i=0; i<nofAttributes; i++) {
				(*spectralCoefficient)[i] += fabs((*spectralCoefficient)[i])*factor;
			}
		}		
		
		// update position of next watermark bit to embed
		bitPosition++;
		if (bitPosition>=160) {
			// automatically wrap around and embed the first bit again
			bitPosition=0;
		}
	}
	
	// process backward/inverse transformation
	statusBar->showMessage ("Backward Transformation...", 0);
	frequencyTransformation->backwardTransformation(patch, spectralList, dataPack);
	
	// free memory
	delete frequencyTransformation;
}


/**
 * Extracts the public watermark by using the test geometry.
 * Since embedding is dependent on a passphrase, the same passphrase has to be used for extraction.
 *
 * @param passPhrase	The passphrase to be used for extraction.
 * @param watermark		A watermark configured with the options the user specified. Note that the 
 *						watermark will be filled with content later on.
 */
void PublicWatermarking::extractWatermark(const QString& passPhrase, WatermarkInterface* originalWatermark) {

	// declare and define some data structures
	StatusBar                *statusBar;
	SurfelInterface          *surfel;
	SurfelCollection         *activeSurfelCollection;
	Scene                    *scene;
	uint                     nofSurfels;
	uint					 i;
	QVector<SurfelInterface> *selection;
	SelectionTool            *selectionTool;
	statusBar    = StatusBar::getInstance();
	scene        = Scene::getInstance();

	// inform user of the start of the extraction process
	statusBar->resetProgress();
	statusBar->showMessage ("Extracting Watermark...", 0);

	// get the surfels of the active object
	activeSurfelCollection = Scene::getInstance()->getActiveObject()->getSurfelCollection();

	// create list of surfels
	MySurfelList surfelList;

	// two different cases (selection, noselection)
	selectionTool = CoreTools::getInstance()->getSelectionTool();
	if (selectionTool->hasSelection() == false) {

		// the user hasn't selected any surfels, so let's use the active object
		// also determine the number of surfels contained in the active object
		nofSurfels = activeSurfelCollection->getNofSurfels();

		// create list of surfels
		for (i = 0, surfel = activeSurfelCollection->getFirstSurfel(); i < nofSurfels; i++, surfel = activeSurfelCollection->getNextSurfel()) {
			surfelList.append(surfel);
		}
	} else {

		// the user has selected some surfels, so let's use these for the extraction process
		selection = selectionTool->getSelection();
		nofSurfels = selection->size();

		// create list of surfels
		for (i = 0; i < nofSurfels - 1; i++) {
			surfel = selection->at(i);
			surfelList.append(surfel);
		}
	}

	// preprocessing: reorientation using PCA
	// the generated secret seed must depend on some features of the original object,
	// i.e. eigenvalues of PCA for instance. this prevents some forgery attacks.
	// the same feature has been extracted while embedding the watermark
	PreprocessingInterface* preprocessor = new PreprocessingImplementation();
	char* feature = preprocessor->extractFeature(&surfelList); 
	preprocessor->preprocess(&surfelList);

	// the feature we have created is a dummy one (its value has no meaning because it has
	// been created using the modified file and not the original one.
	// we have to restore the original feature now.
	FILE * file = fopen("feature.txt", "r+");
	fgets(feature, 250, file);
	fclose(file);

	// inform the user that the feature has been restored
	QMessageBox::information(NULL, "Feature Restored", feature);

	// create the watermark bit string
	// process passphrase (first SHA-1, then CRC to seed the RNG)
	SHA1 sha1;
	sha1.reset();
	sha1.update((unsigned char *)(const char *)passPhrase, passPhrase.length());
	sha1.update((unsigned char *)(const char *)feature, strlen(feature));
	sha1.final();
	unsigned char hashValue[20];
	sha1.getHash(hashValue);

	// note that this watermark has to be the same type as the watermark embedded into
	// the object because it will be passed to some of the following routines to 
	// initialize data correctly (watermark capacity determines for instance the size 
	// of the patches).
	// generate the watermark. as you can see it depends on the feature of the original 
	// object and the passphrase (both are needed).
	for (unsigned int bitNumber=0; bitNumber<originalWatermark->getCapacityInBits(); bitNumber++) {
		originalWatermark->setBit(bitNumber, hashValue[bitNumber/8]&(1<<(bitNumber%8)));		
	}

	// we want to make the random number generator dependent on the previously calculated
	// secret hash value. unfortunately 160 bit is too long, so we use a simple CRC32 
	// routine to map the watermark value to 32 bit which corresponds to a double value.
	// this can be used to seed the PRNG.
	CRC32 crc32;
	double seed = crc32.calcCRC(hashValue, 20);
	
	// intermediary clean up
	delete preprocessor; 
	delete[] feature;

	// get the patches here using PCA space subdivision.
	// note that the patch generator is deterministic although it contains a random number generator.
	// this is because the PRNG is always initialized to the same value. because we use the original
	// object again, we will get the same patches we got when embedding the watermark
	PatchBuilderInterface* patchBuilder = new PatchBuilderImplementation(originalWatermark);
	PatchBuilderInterface::PatchList* patchList = patchBuilder->buildPatches(&surfelList);
	delete patchBuilder; 

	// extract watermark correlation and threshold from every patch
	QList<float> correlationList;
	QList<float> thresholdList;

	for (Patch* patch = patchList->first(); patch!=0; patch = patchList->next()) {
		// initialize random number generator with a secret value
		delete random;
		random = new Random(seed);
		float* currentCorrelation = new float();
		float* currentThreshold = new float();
		extractWatermarkFromPatch(patch, originalWatermark, currentCorrelation, currentThreshold);
		qDebug("Correlation: %f / Threshold: %f\r\n", *currentCorrelation, *currentThreshold);
		correlationList.append(currentCorrelation);
		thresholdList.append(currentThreshold);
	}

	// inform user and display dialog box
	statusBar->setProgress (100, true);
	statusBar->showMessage ("", 1);
	CorrelationDialogImpl correlationDialog(&correlationList, &thresholdList, 0, "Watermark Correlation for every Patch", true);
	correlationDialog.show();
	correlationDialog.raise();

	// should clean up lists
	correlationList.clear();
	thresholdList.clear();

	/*
	// we now want to create some random watermark values and check for their correlation
	// their correlation should always be below threshold
	// extract watermark from every patch
	QList<float> correlationTestList;
	QList<float> thresholdTestList;
	
	// since canvas is 470 pixels wide, we generate so many random combinations
	patch = patchList->first();
	for (i=0; i<470; i++) {
		// create test watermark
		uchar testWatermark[20];
		for (int i=0; i<20; i++) {
			testWatermark[i] = (unsigned char)(random->getUniform()*256);
		}
		float* currentCorrelation = new float();
		float* currentThreshold   = new float();
		extractWatermarkFromPatch(patch, testWatermark, currentCorrelation, currentThreshold);
		correlationTestList.append(currentCorrelation);
		thresholdTestList.append(currentThreshold);
		patch = patchList->next();
		if (patch==0) {
			patch = patchList->first();
		}
	}
	CorrelationDialogImpl testCorrelationDialog(&correlationTestList, &thresholdTestList, 0, "Random Watermarks Correlation", false);
	testCorrelationDialog.show();
	testCorrelationDialog.raise();

	// should clean up lists
	correlationTestList.clear();
	thresholdTestList.clear();
	*/

	// print message for the user
	QMessageBox::information( NULL, "Information", "Watermark has been extracted.\nResults can be seen in the graphs." );
}


/**
 * Extracts watermark from a patch. In fact, no watermark is explicitly extracted, but a correlatiion and a threshold value
 * is calculated.
 *
 * @param patch						The patch to be processed.
 * @param watermark					The watermark that is supposed to be contained in the patch
 * @param correlation				This will contain the resulting correlation.
 * @param threshold					This will contain the resulting threshold.
 */
void PublicWatermarking::extractWatermarkFromPatch(Patch* patch, WatermarkInterface* watermark, float* correlation, float* threshold) {

	uint i;

	// init transformation data structure for the points of patch
	FTInterface* frequencyTransformation = new	FTImplementation(watermark);
	FTInterface::SpectralList* spectralList; // this will contain the spectral components of the transformed object
	FTInterface::DataPack* dataPack;		 // this will contain some temporary data used for backward transformation

	// process forward transformation
	frequencyTransformation->forwardTransformation(patch, &spectralList, &dataPack);
	
	// reset correlation and threshold values
	*correlation = 0.0f;
	*threshold = 0.0f;

	// calculate the correlation and the threshold
	int bitPosition = 0;
	int count = 0;
	for (Spectrum* spectralCoefficient = spectralList->first(); spectralCoefficient!=0; spectralCoefficient = spectralList->next()) {
	
		float watermarkBit = random->getGaussian();
		if (watermark->getBit(bitPosition)==0) {
			watermarkBit = -watermarkBit;
		}		
		for (i=0; i<nofAttributes; i++) {
			*correlation += (*spectralCoefficient)[i]*watermarkBit;
			*threshold += fabs((*spectralCoefficient)[i]);
		}
		count += nofAttributes;
				
		// update position of next watermark bit to embed
		bitPosition++;
		if (bitPosition>=160) {
			// automatically wrap around and restart with first bit
			bitPosition=0;
		}
	}

	// calculate final correlation and treshold
	*correlation *= 1.0f/(count);
	*threshold *= 0.2/(3*count);
	
	// free memory
	delete frequencyTransformation;
}

/**
 * Returns the name of this module.
 *
 * @return The name of this module.
 */
const char* PublicWatermarking::getModuleName() const {
	return "Public Watermark";
}
  
/**
 * Returns the description of this module.
 *
 * @return The description of this module.
 */
const char* PublicWatermarking::getModuleDescription() const {
	return	"This algorithm implements a public watermark.\n"
			"Verification doesn't require the original object.\n"
			"Embedding is performed by modifying the spectral\n"
			"coefficients computed by solving an eigenproblem\n"
			"on the laplacian matrix of the points.";
}

/**
 * Test routine for evaluating the feasability of the chosen approach based on correlation.
 *
 * Das Resultat zeigt, dass das oeffentliche watermarking-verfahren in dieser form nicht implementiert werden
 * kann, da die korrelationsberechnungen nur sinnvolle resultate liefern ab einer laenge von 16000.
 * einige hundert frequenzwerte, wie wir sie bei einer watermark-laenge von 160 bit haben, reichen 
 * nicht (es gibt zu viele zufaellige watermarks, die eine korrelation liefern, die ueber dem
 * threshold liegt). ausserdem ist in einigen faellen die korrelation des richtigen watermarks unter
 * dem threshold. dies ist aber nicht weiter tragisch, da einfach ein anderer watermark-wert generiert
 * werden koennte.
 *
 * fuer die werte len = 16000 und alpha = 0.1f ergeben sich folgende resultate:
 * 
 * Correlation: 4.455629
 * Threshold: 1.667370
 * Now using random watermarks:
 * Failure Percentage: 0.00 
 *
 * Correlation: 5.013888
 * Threshold: 1.661320
 * Now using random watermarks:
 * Failure Percentage: 0.00 
 *
 * Correlation: 5.190956
 * Threshold: 1.658031
 * Now using random watermarks:
 * Failure Percentage: 0.00 
 *
 * Correlation: 5.041628
 * Threshold: 1.663743
 * Now using random watermarks:
 * Failure Percentage: 0.00 
 *
 * Correlation: 4.208990
 * Threshold: 1.664785
 * Now using random watermarks:
 * Failure Percentage: 0.00 
 *
 * Correlation: 4.745228
 * Threshold: 1.667690
 * Now using random watermarks:
 * Failure Percentage: 0.00 
 * 
 * Correlation: 4.800943
 * Threshold: 1.669113
 * Now using random watermarks:
 * Failure Percentage: 0.00 
 *
 * hier scheint das ganze sehr zuverlaessig zu funktionieren.
 *
 * @return Always 0; 
 */
int debug() {

	// define length of the watermark
	int len = 50000;

	// define amplitude of modification
	float alphax = 0.1f;

	// initialize random number generator
	Random random(3615);


	// generate a random watermark of the required size
	WatermarkInterface* watermark = new MajorityWatermark(len, 0.01f, 3, 10, false, false);
	for (int i=0; i<len; i++) {
		watermark->setBit(i, random.getUniform()<0.5f);
	}

	// generate a random spectrum large enough to contain all the watermark bits
	Vector3D* spectrum = new Vector3D[len];
	for (i=0; i<len; i++) {
		spectrum[i][0] = 1+random.getUniform()*50;
		spectrum[i][1] = 1+random.getUniform()*50;
		spectrum[i][2] = 1+random.getUniform()*50;
	}

	// calculate a random number generator seed value based on the watermark	
	CRC32 crc32;
	double seed = crc32.calcCRC((unsigned char *)(*watermark), 20);

	// initialize a random number generator with the previously calculated seed value
	// we will use this generator for the embedding and the extraction process
	Random randomEmbedding(seed);

	// embed watermark
	for (i=0; i<len; i++) {
		float factor = alphax*randomEmbedding.getGaussian();
		spectrum[i][0] += fabs(spectrum[i][0])*factor;
		spectrum[i][1] += fabs(spectrum[i][1])*factor;
		spectrum[i][2] += fabs(spectrum[i][2])*factor;
	}

	// reset the correlation and the threshold value for the extraction
	float correlation = 0.0f;
	float threshold = 0.0f;

	// we need to initialize the random number generator with the same seed that 
	// has been used during embedding
	Random randomExtraction(seed);

	// calculate the correlation and the threshold
	int count = 0;
	for (i=0; i<len; i++) {
		float watermarkBit = randomExtraction.getGaussian();
		correlation += spectrum[i][0]*watermarkBit;
		correlation += spectrum[i][1]*watermarkBit;
		correlation += spectrum[i][2]*watermarkBit;
		count += 3;
		threshold += fabs(spectrum[i][0]);
		threshold += fabs(spectrum[i][1]);
		threshold += fabs(spectrum[i][2]);
	}

	// calculate final values of the correlation and the threshold
	correlation *= 1.0f/(count);
	threshold *= alphax/(3*count);

	// output information to the user
	qDebug("Correlation: %f\nThreshold: %f", correlation, threshold);
	qDebug("Now using random watermarks:");

	// this will count how many of the randomly generated watermark results
	// in a correlation above the threshold. we want to keep this number
	// as small as possible
	int countWarning = 0;

	// let's generate 1000 random watermarks
	for (int j=0; j<1000; j++) {

		// generate watermark
		for (int i=0; i<len; i++) {
			watermark->setBit(i, random.getUniform()<0.5f);
		}

		// initialize the random number generator with the seed which depends on the watermark
		double seed = crc32.calcCRC((unsigned char *)(*watermark), 20);
		Random r(seed);

		// reset the correlation and the threshold value for the extraction
		correlation = 0.0f;
		threshold = 0.0f;
	
		// calculate the correlation and the threshold
		count = 0;
		for (i=0; i<len; i++) {
			float watermarkBit = r.getGaussian();
			correlation += spectrum[i][0]*watermarkBit;
			correlation += spectrum[i][1]*watermarkBit;
			correlation += spectrum[i][2]*watermarkBit;
			count += 3;
			threshold += fabs(spectrum[i][0]);
			threshold += fabs(spectrum[i][1]);
			threshold += fabs(spectrum[i][2]);
		}

		// calculate final values of the correlation and the threshold
		correlation *= 1.0f/(count);
		threshold *= alphax/(3*count);

		// if the correlation is above the threshold, then issue a warning and increment the counter.
		if (correlation>threshold) {
			//qDebug("---->>>> WARNING:");
			//qDebug("     Correlation: %02.6f\n     Threshold: %02.6f\n", correlation, threshold);
			countWarning++;
		} else {
			//qDebug("Correlation: %02.6f\nThreshold:   %02.6f", correlation, threshold);
		}
	}

	// we no longer need the original watermark, so let's delete it
	delete watermark;

	// output the information about the failure percentage and return 
	qDebug("Failure Percentage: %3.6f \n", countWarning/10.00f);
	return 0;
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
