//
// Title:   Random.cpp
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

/**
 * @file Random.cpp
 * @brief Contains code for a random number generator based on Knuth and 
 *        code from www.nr.com.
 *
 * Currently uniform random numbers and gaussian random numbers are implemented.
 *
 * @author Daniel Cotting
 */

#include <math.h>
#include <stdlib.h> 
#include "Random.h"

// According to Knuth, any large MBIG, and any smaller (but still large) MSEED 
// can be substituted for the values below.
#define MBIG 1000000000
#define MSEED 161803398
#define MZ 0
#define FAC (1.0/MBIG)

/** 
 * Constructor for the random number generator class.
 *
 * @param seed Value used to seed the random number generator.
 */
Random::Random(long seed) {

	// seed the random number generator
	long mj,mk;
	int i,ii,k;
	mj=labs(MSEED-labs(seed)); 	    // Initialize ma[55] using the seed idum and the
									// large number MSEED. 
	mj %= MBIG;
	ma[55]=mj;
	mk=1;
	for (i=1;i<=54;i++) {			// Now initialize the rest of the table,
		ii=(21*i) % 55; 			// in a slightly random order,
		ma[ii]=mk; 					// with numbers that are not especially random.
		mk=mj-mk;
		if (mk < MZ) mk += MBIG;
		mj=ma[ii];
	}
	for (k=1;k<=4;k++) 				// We randomize them by warming up the generator.
		for (i=1;i<=55;i++) {
			ma[i] -= ma[1+(i+30) % 55];
			if (ma[i] < MZ) ma[i] += MBIG;
		}
	inext=0; 						// Prepare indices for our first generated number.
	inextp=31; 						// The constant 31 is special; see Knuth.
}

/**
 * Calculates a uniformly distributed pseudo random number.
 *
 * Code is based on algorithm presented in numerical recipes, chapter 7.1. (see www.nr.com)
 * 
 * @return A uniform random deviate between 0.0 and 1.0.
 */
double Random::getUniform() {
	long mj;
	// Here is where we start, except on initialization.
	if (++inext == 56) inext=1; 	// Increment inext and inextp, wrapping around 56 to 1. 
	if (++inextp == 56) inextp=1;
	mj=ma[inext]-ma[inextp]; 		// Generate a new random number subtractively.
	if (mj < MZ) mj += MBIG; 		// Be sure that it is in range.
	ma[inext]=mj; 					// Store it,
	return mj*FAC; 					// and output the derived uniform deviate.
}

/**
 * Calculates a normally distributed pseudo random number.
 *
 * Code is based on algorithm of box-muller presented in numerical recipes,
 * chapter 7.2. (see www.nr.com)
 *
 * @return A gaussian random number.
 */
double Random::getGaussian()
{
	static int iset=0;
	static double gset;
	double fac,rsq,v1,v2;
	if (iset == 0) {							// We don't have an axtra deviate handy, so
		do {
			v1=2.0*this->getUniform()-1.0;	    // pick two uniform numbers in the square
			v2=2.0*this->getUniform()-1.0;	    // extending from -1 to 1 in each direction,
			rsq=v1*v1+v2*v2;					// see if thez are in the unit circle,
		} while (rsq >= 1.0 || rsq == 0.0);		// and if thez are not, try again.
		fac=sqrt(-2.0*log(rsq)/rsq);
		// Now make the Box-Muller transformation to get two normal deviates.
		// Return one and save the other for next time.
		gset=v1*fac;
		iset=1;									// set the flag
		return v2*fac;
	} else {									// We have an extra deviate handy,
		iset=0;									// so unset the flag,
		return gset;			 				// and return it.
	}
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
