//
// Title:   SteganoWatermarking.cpp
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

/**
 * @file SteganoWatermarking.cpp
 *
 * This file implements a steganography algorithm which allows to embed a text into 
 * a 3D geometry.
 */
#include <qobject.h>
#include <qmessagebox.h>
#include <assert.h>
#include <time.h>

#include "../../../Core/Scene/src/Scene.h"
#include "../../../Core/Scene/src/Object.h"
#include "../../../Core/DataStructures/src/SurfelCollection.h"
#include "../../../Core/CoreTools/src/CoreTools.h"
#include "../../../Core/CoreTools/SelectionTool/src/SelectionTool.h"
#include "../../../Utilities/UserMessages/src/StatusBar.h"
#include "../../../Core/Interfaces/src/SurfelInterface.h"

#include "MySurfelList.h"
#include "Patch.h"
#include "SteganoWatermarking.h"
#include "WatermarkingInterface.h"
#include "MajorityWatermark.h"
#include "PatchBuilderInterface.h"
#include "RegistrationInterface.h"
#include "PCARegistration.h"
#include "PreprocessingInterface.h"
#include "PCAPreprocessing.h"
#include "PCAPatchBuilder.h"
#include "Random.h"
#include "SHA1.h"
#include "CRC32.h"
#include "LaplaceFrequencyTransformation.h"
#include "FrequencyTransformationInterface.h"
#include "CorrelationDialogImpl.h"

// the implementation of various algorithms can be changed by adapting the following typedefs
typedef MajorityWatermark WatermarkImplementation;						// class holding the watermark
typedef PCAPreprocessing PreprocessingImplementation;					// preprocessing applied to the surfels to be watermarked
typedef PCAPatchBuilder PatchBuilderImplementation;						// class used for building the patches 

// parameters for the watermarking follow:
// specifies how many attribute values per surfel we want to use for the watermarking process
// usually we will use 3 of them: x, y and z coordinate.
#define nofAttributes 3

// some typedefs for the frequency transformation:
// to replace the laplace frequency transformation by another algorithm, the following
// lines can be adapted.
typedef FrequencyTransformationInterface<nofAttributes> FTInterface;	// frequency transformation interface, convenience short version
typedef GenericVector<nofAttributes> Spectrum;							// spectrum typedef, convenience short version
typedef LaplaceFrequencyTransformation<nofAttributes> FTImplementation;	// frequency transformation implementation

/**
 * Embeds a watermark into the currently selected surfels or the active object.
 * The watermark is based on a passphrase, which is passed to this function.
 * Although a selection is supported in the embedding process, the extraction procedure
 * only supports PointShop3D objects and no selection. Therefore selections have to be converted 
 * to objects before extraction.
 *
 * @param passPhrase	The secret, user defined passphrase to use for the watermark creation.
 * @param watermark		A watermark configured with the options the user specified. Note that the 
 *						watermark will be filled with content later on.
 */
void SteganoWatermarking::embedWatermark(const QString& passPhrase, WatermarkInterface* watermark) 
{
	// some data structures used for the embedding process
	StatusBar                *statusBar;
	SurfelInterface          *surfel;
	SurfelCollection         *activeSurfelCollection;
	Scene                    *scene;
	uint                     nofSurfels;
	uint					 i;
	QVector<SurfelInterface> *selection;
	SelectionTool            *selectionTool;
	statusBar    = StatusBar::getInstance();
	scene        = Scene::getInstance();

	// get the surfels of the currently active object
	activeSurfelCollection = Scene::getInstance()->getActiveObject()->getSurfelCollection();

	// reset the user information, i.e. set the progress bar to zero and show the status bar message
	statusBar->resetProgress();
	statusBar->showMessage ("Starting Embedding Watermark...", 0);

	// this will contain the list of surfels to modify 
	MySurfelList surfelList;

	// creation of the list has two different cases (selection, noselection)
	statusBar->showMessage ("Creating Surfel Lists...", 0);
	selectionTool = CoreTools::getInstance()->getSelectionTool();
	if (selectionTool->hasSelection() == false) {

		// the user hasn't selected any surfels. we will therefore use the active object
		// get the number of surfels we have to add to the list
		nofSurfels = activeSurfelCollection->getNofSurfels();
		// create list of surfels
		for (i = 0, surfel = activeSurfelCollection->getFirstSurfel(); i < nofSurfels; i++, surfel = activeSurfelCollection->getNextSurfel()) {
			surfelList.append(surfel);
		}

	} else {

		// the user has selected a portion of the surfels
		selection = selectionTool->getSelection();
		// get the number of surfels we have to add to the list
		nofSurfels = selection->size();
		// create list of surfels
		for (i = 0; i < nofSurfels - 1; i++) {
			surfel = selection->at(i);
			surfelList.append(surfel);
		}
	}

	// create the patches
	// the current implementation uses PCA to subdivide the space recursively
	// the patches have to be created before modifying the original object
	statusBar->showMessage ("Creating Patches...", 0);
	PatchBuilderInterface* patchBuilder = new PatchBuilderImplementation(watermark);
	PatchBuilderInterface::PatchList* patchList = patchBuilder->buildPatches(&surfelList);
	delete patchBuilder; 

	// check whether the user wants to color the patches (used for debugging and education)
	switch( QMessageBox::information( NULL, "Question",	"Do you want to color the patches?",
								"Yes", "No", 0, 0, 1 ))	{
	case 0: 
		// Define a random number generator for the coloring of the patches
		Random randomForColor(3615);

		// now process every surfel of the object and color it
		for (Patch* colorPatch=patchList->first(); colorPatch!=0; colorPatch=patchList->next()) {
			
			// define the color of the current patch
			QColor color; 
			float red   = randomForColor.getUniform() * 256;
			float green = randomForColor.getUniform() * 256; 
			float blue  = randomForColor.getUniform() * 256;
			color.setRgb(red, green, blue);

			// apply the color to all the surfels of the current patch
			for (SurfelInterface* surfel=colorPatch->first(); surfel!=0; surfel=colorPatch->next()) {
				// modify the color
				scene->setSurfelDiffuseColor(surfel, color.rgb(), false);
			}
		}
		
		// emit the signal in order to update the screen 
		colorPatch=patchList->first();
		SurfelInterface* firstSurfel=colorPatch->first();
		QColor color = firstSurfel->getDiffuseColor();
		scene->setSurfelDiffuseColor (firstSurfel, color.rgb(), true);
	}

	statusBar->showMessage ("Preprocessing...", 0);
	PreprocessingInterface* preprocessor = new PreprocessingImplementation();
	// extract features from the point cloud (watermark has to depend on original object)
	// this task is also performed by the preprocessor
	char* feature = preprocessor->extractFeature(&surfelList); 
	delete preprocessor; 

	// create the watermark bit string by running SHA-1 on passphrase and feature
	statusBar->showMessage ("Generating Secure Watermark...", 0);
	SHA1 sha1;
	sha1.reset();
	sha1.update((unsigned char *)(const char *)passPhrase, passPhrase.length());
	sha1.update((unsigned char *)(const char *)feature, strlen(feature));
	sha1.final();
	unsigned char hashValue[21]; 
	sha1.getHash(hashValue);
	hashValue[20]=0;
	qDebug("Hash Value at Embedding: %s", hashValue);

	// we no longer need the feature, so clean up
	delete[] feature;

	// propagate the passphrase value to the watermark
	for (unsigned int bitNumber=0; bitNumber<watermark->getCapacityInBits(); bitNumber++) {
		// check if passphrase is large enough
		if (bitNumber/8<passPhrase.length()) {
			watermark->setBit(bitNumber, ((const char*)passPhrase)[bitNumber/8]&(1<<(bitNumber%8)));
		} else {
			watermark->setBit(bitNumber, false);
		}
	}

	// calculate the seed for the random number generator
	CRC32 crc32;
	double seed = crc32.calcCRC(hashValue, 20);

	// we will output some timing information for the following code
	char statusMessage[20];
	FILE* file = fopen("log.txt", "a+");
	assert(file!=NULL);

	// apply the watermark to every patch
	uint currentPatchNumber=0;
	uint totalPatchNumber=patchList->count();
	for (Patch* patch = patchList->first(); patch!=0; patch = patchList->next()) {

		// initialize random number generator with a secret value
		delete random;
		random = new Random(seed);

		// display process status
		sprintf(statusMessage, "Patch #%04d", ++currentPatchNumber);
		statusBar->showMessage (statusMessage, 0);
		statusBar->setProgress ((int)(currentPatchNumber/(float)totalPatchNumber*100.0), true);

		// initialize timing data structures
		clock_t tv1, tv2;
		double time;
		tv1 = clock();

		// process current patch
		embedWatermarkIntoPatch(patch, watermark);

		// calculate the time elapsed during the processing of the patch 
		// and output the information to the file and flush it
		tv2 = clock();
		time = (tv2 - tv1)/((double)CLOCKS_PER_SEC);
		// fprintf(file, "Timing: %f s for patch with %u points and %u eigenvalues.\n", time, patch->count(), watermark->getUsedSizeInBits());
		// fflush(file);
	}

	// clean up, i.e. close timing information file and delete the list of patches
	fclose(file);
	delete patchList;

	// we emit the signal in order to update the screen
	statusBar->showMessage ("Updating Screen...", 0);
	surfel = activeSurfelCollection->getFirstSurfel();
	scene->setSurfelPosition (surfel, surfel->getPosition(), true);

	// show message dialog box to inform the user of task completion
	QMessageBox::information( NULL, "Information", "Watermark has been embedded.\nYou can now save your watermarked object." );

	// reset status bar message and set progress bar to 100%
	statusBar->setProgress (100, true);
	statusBar->showMessage ("", 1);
}

/**
 * Embeds the watermark into the patch.
 *
 * @param patch The patch to embed the watermark into.
 * @param watermark The watermark to embed into the patch.
 */
void SteganoWatermarking::embedWatermarkIntoPatch(Patch* patch, WatermarkInterface* watermark) 
{

	uint i;
	StatusBar* statusBar = StatusBar::getInstance();

	// init transformation data structure
	FTInterface* frequencyTransformation = new	FTImplementation(watermark);	// the algortihm object for transformation
	FTInterface::SpectralList* spectralList;									// will contain the result of the transformation
	FTInterface::DataPack* dataPack;											// contains additional information allowing backward transformation
	
	// process forward transformation
	statusBar->showMessage ("Forward Transformation...", 0);
	frequencyTransformation->forwardTransformation(patch, &spectralList, &dataPack);
	
	// embed watermark into the frequencies
	statusBar->showMessage ("Embedding...", 0);
	unsigned int bitPosition = 0;						// which bit do we have to embed into the current frequency?
	int spectrumPosition = 0;							// which spectral coefficient are we processing now?
	float watermarkStrength = watermark->getStrength(); // strength of frequency modulation

	// embed data into all the spectral values which have been computed
	// note that more than watermark->getCapacity() bits can be embedded if the chip-rate is set 
	// to a higher value than 1. in that case we will simply wrap around and restart with the 
	// embedding of the first bit after having embedded the last one.
	qDebug("---------->");
	for (Spectrum* spectralCoefficient = spectralList->first(); spectralCoefficient!=0; spectralCoefficient = spectralList->next()) {

		// modify spectrum depending on watermark bit.
		if (watermark->getBit(bitPosition)) {
			// watermark bit is set to 1
			// embed the same bit into all the three attributes
			for (i=0; i<nofAttributes; i++) {
				// make a certain digit odd
				float value = (*spectralCoefficient)[i];
				qDebug("%f", value);
				int intValue = (int)floor(value);
				if (intValue&1) {
					(*spectralCoefficient)[i] += 1.f;
				}
			}
		} else {
			// watermark bit is set to 0
			// embed the same bit into all the three attributes
			for (i=0; i<nofAttributes; i++) {
				// make a certain digit even
				float value = (*spectralCoefficient)[i];
				qDebug("%f", value);
				int intValue = (int)floor(value);
				if (intValue&1) {
					(*spectralCoefficient)[i] += 1.f;
				}
			}
		}		

		// update the position of the next spectral coefficient
		spectrumPosition++;

		// update position of next watermark bit to embed
		bitPosition++;
		if (bitPosition>=watermark->getCapacityInBits()) {
			// automatically wrap around and embed the first bit again
			bitPosition=0;
		}
	}


	// process backward/inverse transformation
	statusBar->showMessage ("Backward Transformation...", 0);
	// note that the spectral list and the data pack are input parameters this time
	frequencyTransformation->backwardTransformation(patch, spectralList, dataPack);

	// free memory
	delete frequencyTransformation;
}

/**
 * Extracts the private watermark by using the test object and the original object. Note that
 * both geometries have to be available as objects. No selection is supported during extraction.
 * That's the place where our private watermarking algorithm is significantly different from the 
 * public one (here we need two objects: the original one and the object to be verified).
 * Since embedding is dependent on a passphrase, the same passphrase has to be used for extraction.
 *
 * @param passPhrase			The passphrase to be used for extraction.
 * @param originalWatermark		A watermark configured with the options the user specified. Note that the 
 *								watermark will be filled with content later on.
 */
void SteganoWatermarking::extractWatermark(const QString& passPhrase, WatermarkInterface* originalWatermark) {

	uint					 i;
	StatusBar                *statusBar;
	SurfelInterface          *surfel;
	Scene                    *scene;

	Object					 *originalObject;
	SurfelCollection         *originalSurfelCollection;
	uint                     nofOriginalSurfels;

	// show message in status bar
	statusBar = StatusBar::getInstance();
	statusBar->resetProgress();
	statusBar->showMessage ("Extracting Watermark...", 0);

	// get instance of the scene for later usage
	scene = Scene::getInstance();

	// get two objects from the scene
	Object* firstObject = scene->getFirstObject();

	originalObject = firstObject;

	// get the collection of surfels of both objects
	originalSurfelCollection = originalObject->getSurfelCollection();

	// create list of surfels
	// note that in this watermarking scheme we have to create two sets of surfels
	// one set contains the original object, whereas the second one contains to object to be verified
	statusBar->showMessage ("Creating Surfel List...", 0);
	
	// first create the surfel list of the original object
	MySurfelList originalSurfelList;
	nofOriginalSurfels = originalSurfelCollection->getNofSurfels();
	for ( i = 0, surfel = originalSurfelCollection->getFirstSurfel(); 
		  i < nofOriginalSurfels; 
		  i++, surfel = originalSurfelCollection->getNextSurfel()) 
	{
		originalSurfelList.append(surfel);
	}
	
	// get the patches here using PCA space subdivision.
	// note that the patch generator is deterministic although it contains a random number generator.
	// this is because the PRNG is always initialized to the same value. because we use the original
	// object again, we will get the same patches we got when embedding the watermark
	// patches have to be generated before modifying and preprocessing the original object
	statusBar->showMessage ("Creating Patches...", 0);
	PatchBuilderInterface* patchBuilder = new PatchBuilderImplementation(originalWatermark);
	PatchBuilderInterface::PatchList* originalPatchList = patchBuilder->buildPatches(&originalSurfelList);
	delete patchBuilder; 

	// check whether the user want to color the patches (used for debugging and education)
	switch( QMessageBox::information( NULL, "Question",	"Do you want to color the patches?",
								"Yes", "No", 0, 0, 1 ))	{
	case 0: 
		// Define a random number generator for the coloring of the patches
		Random randomForColor(3615);

		// now process every surfel of the object and color it
		for (Patch* colorPatch=originalPatchList->first(); colorPatch!=0; colorPatch=originalPatchList->next()) {
			
			// define the color of the current patch
			QColor color;  
			float red   = randomForColor.getUniform() * 256;
			float green = randomForColor.getUniform() * 256; 
			float blue  = randomForColor.getUniform() * 256;
			color.setRgb(red, green, blue);

			// apply the color to all the surfels of the current patch
			for (SurfelInterface* surfel=colorPatch->first(); surfel!=0; surfel=colorPatch->next()) {
				// modify the color
				scene->setSurfelDiffuseColor(surfel, color.rgb(), false);
			}
		}
		
		// emit the signal in order to update the screen 
		colorPatch=originalPatchList->first();
		SurfelInterface* firstSurfel=colorPatch->first();
		QColor color = firstSurfel->getDiffuseColor();
		scene->setSurfelDiffuseColor (firstSurfel, color.rgb(), true);
		break;
	}

	// the generated secret seed must depend on some features of the original object,
	// i.e. eigenvalues of PCA for instance. this prevents some forgery attacks.
	// the same feature has been extracted while embedding the watermark
	// note that the feature has to be extracted before any modification on the 
	// original object is performed (e.g. registration)
	statusBar->showMessage ("Extracting Feature...", 0);
	PreprocessingInterface* preprocessor = new PreprocessingImplementation();
	char* feature = preprocessor->extractFeature(&originalSurfelList); 
	// show feature dialog box
	// QMessageBox::information(NULL, "Extracted Feature", feature);
	delete preprocessor; 

	// create the watermark bit string
	// process passphrase (first SHA-1, then CRC to seed the RNG)
	SHA1 sha1;
	sha1.reset();
	sha1.update((unsigned char *)(const char *)passPhrase, passPhrase.length());
	sha1.update((unsigned char *)(const char *)feature, strlen(feature));
	sha1.final();
	unsigned char hashValue[20];
	sha1.getHash(hashValue);
	hashValue[20]=0;
	qDebug("Hash Value at Extraction: %s", hashValue);
	delete[] feature;

	// note that this watermark has to be the same type as the watermark embedded into
	// the object because it will be passed to some of the following routines to 
	// initialize data correctly (watermark capacity determines for instance the size 
	// of the patches).
	// final watermark generation step: as you can see it depends
	// on the feature of the original object and the passphrase (both are needed).
	for (unsigned int bitNumber=0; bitNumber<originalWatermark->getCapacityInBits(); bitNumber++) {
		originalWatermark->setBit(bitNumber, false);		
	}

	// we want to make the random number generator dependent on the previously calculated
	// secret hash value. unfortunately 160 bit is too long, so we use a simple CRC32 
	// routine to map the watermark value to 32 bit which corresponds to a double value.
	// this can be used to seed the PRNG.
	CRC32 crc32;
	double seed = crc32.calcCRC(hashValue, 20);
	
	// we will output some status information
	char statusMessage[20];

	// apply the watermark to every patch
	uint currentPatchNumber=0;
	uint totalPatchNumber=originalPatchList->count();

	// extract watermark from every patch
	for (	Patch* patch = originalPatchList->first(); 
			patch!=0; 
			patch = originalPatchList->next()) 
	{

		// display process status
		sprintf(statusMessage, "Patch #%04d", ++currentPatchNumber);
		statusBar->showMessage (statusMessage, 0);
		statusBar->setProgress ((int)(currentPatchNumber/(float)totalPatchNumber*100.0), true);

		// now extract the watermark from this patch
		extractWatermarkFromPatch(patch, NULL, originalWatermark);

	}

	// reset the status bar and progress bar
	statusBar->setProgress (100, true);
	statusBar->showMessage ("", 1);
}

/**
 * Extracts watermark from a patch and a verification object.
 *
 * @param originalPatch				The patch to be processed (on the original object).
 * @param verificationSurfelList	The list containing all the points from the object to be verified.
 * @param originalWatermark			The watermark that is supposed to result from the extraction. It will be
 *									compared with the watermark that is extracted and a correlation value will
 *									be calculated.
 * @return							The correlation value between the specified watermark and the extracted one.
 */
float SteganoWatermarking::extractWatermarkFromPatch(
		Patch* originalPatch,					// patch to process
		MySurfelList* verificationSurfelList,	// project onto this
		WatermarkInterface* originalWatermark)	// compare extracted watermark with this
{

	uint i;
	StatusBar* statusBar = StatusBar::getInstance();

	// init transformation data structure for the points of the original object (patch)
	FTInterface* frequencyTransformation = new	FTImplementation(originalWatermark);
	FTInterface::SpectralList* originalSpectralList; // this will contain the spectral components of the transformed object
	FTInterface::DataPack* originalDataPack;		 // this will contain some temporary data used for backward transformation
	
	// process forward transformation, result will be contained in originalSpectralList and originalDataPack.
	statusBar->showMessage ("Forward Transformation (Original)...", 0);
	frequencyTransformation->forwardTransformation(
		originalPatch, 
		&originalSpectralList, 
		&originalDataPack); 

	float* sumArray = new float[originalWatermark->getCapacityInBits()];

	// intialize with zero values
	for (i=0; i<originalWatermark->getCapacityInBits(); i++) {
		sumArray[i] = 0.0f;
	}


	statusBar->showMessage ("Extracting Watermark...", 0);
	unsigned int bitPosition = 0;				// bit position inside watermark
	Spectrum* originalSpectralCoefficient;		// one spectral coefficient of the original object
	qDebug("----");
	for (originalSpectralCoefficient = originalSpectralList->first();
		 originalSpectralCoefficient!=0;
		 originalSpectralCoefficient = originalSpectralList->next())
	{
		for (i=0; i<nofAttributes; i++) {
			float value = (*originalSpectralCoefficient)[i];
			qDebug("%f", value);
			int intValue = (int)floor(value);
			if ((intValue&1) == 0) {
				sumArray[bitPosition]-=1;
			} else {
				sumArray[bitPosition]+=1;
			}
		}

		// update position of next watermark bit to embed
		bitPosition++;
		if (bitPosition>=originalWatermark->getCapacityInBits()) {
			// automatically wrap around and restart with first bit
			bitPosition=0;
		}
	}

	// write the watermark
	for (i=0; i<originalWatermark->getCapacityInBits(); i++) {
        originalWatermark->setBit(i,sumArray[i]>0);
		qDebug("%d", originalWatermark->getBit(i));
	}	

	char message[100];
	sprintf(message, "Message: %s", (const char*)originalWatermark);
	QMessageBox::information( NULL, "Correlation",	message);

	// no correlation to be calculated
	float correlation = 0.0f;					
	
	delete frequencyTransformation;
	delete originalDataPack;
	delete originalSpectralList;
	delete[] sumArray;

	// return the correlation value
	return correlation;
}

/**
 * Returns the name of this module.
 *
 * @return "Steganography"
 */
const char* SteganoWatermarking::getModuleName() const {
	return "Steganography";
}
  
/**
 * Returns the description of this module.
 *
 * @return The description of this module.
 */
const char* SteganoWatermarking::getModuleDescription() const {
	return	"This algorithm implements a steganography tool.\n"
			"The passphrase is embedded as text into the object.\n"
			"Note that this prototype doesn't include any encryp-\n"
			"tion.";
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
