//
// Title:   WatermarkingInterface.h
// Created: November 28, 2002
// Authors: Daniel Cotting
//
// Copyright (c) 2001, 2002, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

/**
 * @file WatermarkingInterface.cpp
 *
 * Common interface for all watermarking implementations.
 */

#ifndef __WATERMARKINGINTERFACE_H_
#define __WATERMARKINGINTERFACE_H_

#include <qobject.h>
#include "ModuleInterface.h"
#include "WatermarkInterface.h"

/**
 * Common interface for all watermarking implementations.
 */
class WatermarkingInterface: public ModuleInterface {

public:

	/**
	 * Embeds a watermark into the currently selected surfels or the active object.
	 * The watermark is based on a passphrase, which is passed to this function.
	 * Although a selection is supported in the embedding process, the extraction procedure
	 * only supports PointShop3D objects and no selection. Therefore selections have to be converted 
	 * to objects before extraction.
	 *
	 * @param passPhrase	The secret, user defined passphrase to use for the watermark creation.
	 * @param watermark		A watermark configured with the options the user specified. Note that the 
	 *						watermark will be filled with content later on.
	 */
	virtual void embedWatermark(const QString& passPhrase, WatermarkInterface* watermark) = 0;

	/**
	 * Extracts the private watermark by using the test object and the original object. Note that
	 * both geometries have to be available as objects. No selection is supported during extraction.
	 * That's the place where our private watermarking algorithm is significantly different from the 
	 * public one (here we need two objects: the original one and the object to be verified).
	 * Since embedding is dependent on a passphrase, the same passphrase has to be used for extraction.
	 *
	 * @param passPhrase The passphrase to be used for extraction.
	 * @param watermark		A watermark configured with the options the user specified. Note that the 
	 *						watermark will be filled with content later on.
	 */
	virtual void extractWatermark(const QString& passPhrase, WatermarkInterface* watermark) = 0;

	/**
	 * Returns the name of this module.
	 *
	 * @return The name of this module.
	 */
	virtual const char* getModuleName() const = 0;

	/**
	 * Returns the description of this module.
	 *
	 * @return The description of this module.
	 */
	virtual const char* getModuleDescription() const = 0;
};

#endif __WATERMARKINGINTERFACE_H_

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
